/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 *  cache.c
 *
 *  Provide the "cache" resource.  This resource is special since it gets
 *  to look at query responses which are just passing through.
 */

#include "alibi.h"

void
hit_hint ();

float TCF = 1.0;

struct cached_datum {
  struct dobj *datum;
  time_t timestamp;
  long datum_size;
  long total_size;
  long unrealtimecached;
  long hint;
  char *ID;
  char *data_type;
  char *descript;
  struct cached_datum *next;
};

static struct cached_datum *data_cache=NULL, *tail=NULL;

void
slower_turnover ()
{
  TCF += 0.25;
}

void
faster_turnover ()
{
  if (TCF <= 0.0001)
    /* panic ("Higher turnover requested with TCF too low already", 1); */
    return;
  if (TCF <= 0.5)
    TCF /= 2.0;
  else
    TCF -= 0.25;
  return;
}

void
dump_cache (struct conlist *t)
{
  struct cached_datum *a;
  char buf[1000];
  a = data_cache;
  while (a)
  {
    sprintf (buf, "Unreal timestamp %ld real timestamp %s ID %s\n",
      a->unrealtimecached, ctime(&(a->timestamp)), a->ID);
    text_sock_write (t, buf);
    sprintf (buf, "Datum size %ld  Total size %ld\n", a->datum_size,
      a->total_size);
    text_sock_write (t, buf);
    text_sock_write (t, "Data type:  ");
    if (a->data_type)
      text_sock_write (t, a->data_type);
    else
      text_sock_write (t, "(nil)");
    text_sock_write (t, "\nDescription:  ");
    if (a->descript)
      text_sock_write (t, a->descript);
    else
      text_sock_write (t, "(nil)");
    text_sock_write (t, "\n\n");
    a = a->next;
  }
  text_sock_write (t, "Ready.\n");
}

void
free_cached_datum (struct cached_datum *t)
{
  if (!t)
    return;
  free_dobj (t->datum);
  if (t->data_type)
    free (t->data_type);
  if (t->descript)
    free (t->descript);
  free (t);
}

/* Return the amount of memory consumed by a cached_datum.  Assumes that
 * t->datum_size is already set.
 */
long
total_size (struct cached_datum *t)
{
  long count;
  count = sizeof (struct cached_datum);
  if (t->datum)
    count += t->datum_size;
  if (t->data_type)
    count += strlen (t->data_type) +1;
  if (t->descript)
    count += strlen (t->descript) +1;
  return count;
}

/* Return the amount of memory consumed by a datum. */
long
size_dobj (struct dobj *u)
{
  struct dobj *a;
  long count=0;
  a = u;
  while (a)
  {
    count += sizeof (struct dobj);
    if (a->text)
       count += strlen (a->text) +1;
    if (a->bindata)
      count += a->binlen;
    a = a->next;
  }
  return count;
}

/* Add an item to the cache. */
void
add_item (struct cached_datum *t)
{
  struct cached_datum *p, *q;
  freespace -= t->total_size;
  if (freespace < 0)
    panic ("Cache space overrun", 1);
#ifdef debugcache
  printf ("Adding item with unreal TS %ld.\n", t->unrealtimecached);
#endif
  if (!data_cache)
  {
    t->next = NULL;
    data_cache = tail = t;
    return;
  }
  p = data_cache;
  q = NULL;
  while (1)
  {
    if (!p)
      break;
    if (t->unrealtimecached < p->unrealtimecached)
      break;
    q = p;
    p = p->next;
  }
  if (!p)
  {
    if (!q)
      panic ("Logic error in add_item", 1);
    t->next = NULL;
    tail = t;
    q->next = t;
    return;
  }
  t->next = p;
  if (!q)
    data_cache = t;
  else
    q->next = t;
}

/* Remove an item from the cache. */
void
make_room ()
{
  struct cached_datum *t;
  if (!(data_cache))
    panic ("make_room called with empty cache", 1);
  t = data_cache;
  data_cache = t->next;
  if (!(data_cache))
    tail = NULL;
#ifdef debugcache
  printf ("Dumped datum of size %ld with unreal TS %ld.\n", t->total_size,
    t->unrealtimecached);
#endif

  /* Update TCF */
  if (unrealtime - t->unrealtimecached < best_turnover)
    slower_turnover ();
  else if (unrealtime - t->unrealtimecached > best_turnover)
    faster_turnover ();
#ifdef debugcache
  printf ("TCF set to %f\n", TCF);
#endif

  freespace += t->total_size;
  free_cached_datum (t);
}

int
id_check (char *new_id)
{
  struct cached_datum *t;
  t = data_cache;
  while (t)
  {
    if (!(strcmp (t->ID, new_id)))
      return 1;
    t = t->next;
  }
  return 0;
}

struct cached_datum *
kill_old_versions (char *new_id)
{
  struct cached_datum *t, *p;
  if (!new_id)
    return NULL;
  t = data_cache;
  p = NULL;
  while (t)
  {
    if (!(version_blind_id_cmp (t->ID, new_id)))
    {
      unsigned long cache_version, message_version;
      /* They may be identical. */
      if (!(strcmp (t->ID, new_id)))
        return NULL;
      /* Which is newer? */
      sscanf (strrchr (t->ID, ' '), "%lu", &cache_version);
      sscanf (strrchr (new_id, ' '), "%lu", &message_version);
      if (cache_version > message_version)
        return t;

#ifdef debugcache
      printf ("Removing obsolete %s upon receipt of %s\n", t->ID, new_id);
      fflush (stdout);
#endif
      if (p)
        p->next = t->next;
      else
        data_cache = t->next;
      if (!(data_cache))
        tail = NULL;
      freespace += t->total_size;
      free_cached_datum (t);
      return NULL;
    }
    p = t;
    t = t->next;
  }
  return NULL;
}

/*             Hops so far        Running Link Cost     Running Cache Sum
 *    rnd < ----------------- * --------------------- * -----------------
 *          Hops so far + TCF   TCF * SUM (neighbors)          TCF
 */

int
decide (struct dobj *u, long inlag, long dsize, char *new_id)
{
  float fac1, fac2, fac3, rhs;

  /* If already in cache, don't cache it again. */
  if (id_check (new_id))
  {
#ifdef debugcache
    puts ("Item already in cache.");
#endif
    return 0;
  }
  if (!neighborlag)
  {
#ifdef debugcache
    puts ("neighborlag = 0!  (I'm outta here!)");
#endif
    return 0;
  }
  if (sscanf (u->text, "0 Cachedata %lu %lu %lu", &c_hops,
  &c_links, &c_cache) != 3)
  {
#ifdef debugcache
    puts ("Missing cachedata!  (Not gonna cache it)");
#endif
    return 0;
  }
#ifdef debugcache
  printf ("Cachedata from message:  %lu %lu %lu\n", c_hops, c_links, c_cache);
#endif
  c_hops++;
  c_links += inlag;
  c_cache += cachesize;
#ifdef debugcache
  printf ("Updated cachedata:  %lu %lu %lu\n", c_hops, c_links, c_cache);
  printf ("Cache function:         %lu              %lu              %lu\n",
    c_hops, c_links, c_cache);
  printf ("                    ----------    *    ---------    *    ---------\n");
  printf ("                   %lu + %f    %f * %ld       %f\n",
    c_hops, TCF, TCF, neighborlag, TCF);
#endif
  fac1 = (float) c_hops / ((float)c_hops + TCF);
  fac2 = (float) c_links / (TCF * (float)neighborlag);
  fac3 = (float) c_cache / TCF;
  rhs = fac1 * fac2 * fac3;
#ifdef debugcache
  printf ("The factors are:  %f %f %f\n", fac1, fac2, fac3);
  printf ("The product is:  %f\n", rhs);
#endif
  rhs *= fudge_factor;
#ifdef debugcache
  printf ("After fudging, the product is:  %f\n", rhs);
#endif
  if (rnd () < rhs)
  {
#ifdef debugcache
    puts ("Success!");
#endif
    return 2;
  }

#ifdef debugcache
  puts ("rnd was >= rhs.  Too bad.");
#endif

  /* Stage 2 (fill up empty space) */
  /* Estimate extra overhead at 100. */
  if (freespace > dsize + sizeof (struct cached_datum) + 100)
  {
#ifdef debugquery
    puts ("Fortunately, there is enough empty space to cache it anyway!");
#endif
    return 1;
  }

  /* No. */
  return 0;
}

void
remove_one_link (struct dobj **p, struct dobj **u, struct dobj **list)
{
  struct dobj *q;
  q = (*u)->next;
  if (*p)
    (*p)->next = q;
  else
    *list = q;
  if ((*u)->text)
    free ((*u)->text);
  if ((*u)->bindata)
    free ((*u)->bindata);
  free_dobj (*u);
  *u = q;
}

/* This complex function replaces the datum in a response message with a more
 * recent version of the same which was found in cache.  The "enddatum" field
 * of the conlist that eventually owns the message is invalidated here if not
 * already invalid.
 *
 * Only the most recent substitution is noted by the "Upgraded to newer
 * version" header line.
 */
void
kill_message_datum (struct dobj **uu, struct cached_datum *t)
{
  struct dobj *u, *p;
  char buf[namlen+80];
  /* Save only the following fields:
   *   Response path
   *   Query path
   *   Start time
   *   Answer for
   */
  u = *uu;
  p = NULL;
  while (u)
  {
    int flag;
    char junk;
    if (sscanf (u->text, "%d Response path:%c", &flag, &junk) == 2 ||
        sscanf (u->text, "%d Query path:%c", &flag, &junk) == 2 ||
        sscanf (u->text, "%d Start time:%c", &flag, &junk) == 2 ||
        sscanf (u->text, "%d Answer for %c", &flag, &junk) == 2)
    {
      p = u;
      u = u->next;
    }
    else
      remove_one_link (&p, &u, uu);
  }
  /* Tack on cache answer */
  if (!p)
    panic ("Assertion failure in kill_message_datum", 1);
  p->next = clone_dobj (t);
  /* Add the following fields:
   *    Upgraded to newer version
   *    Class
   *    Have a nice day
   */
  add_class (uu, t->data_type);
  while (p->next)
    p = p->next;
  u = (struct dobj *) malloc (sizeof (struct dobj));
  if (!u)
    panic ("Can't malloc in kill_message_datum", 1);
  u->bindata = NULL;
  u->binlen = 0l;
  sprintf (buf, "0 Upgraded to newer version by %s", myhostname);
  u->text = strclone (buf);
  p->next = u;
  p = (struct dobj *) malloc (sizeof (struct dobj));
  if (!p)
    panic ("Can't malloc in kill_message_datum", 1);
  p->bindata = NULL;
  p->binlen = 0l;
  p->text = strclone ("1 Have a nice day =|-)");
  p->next = NULL;
  u->next = p;
}

void
cache_hook (struct dobj **uu, long inlag)
{
  int decision;
  long dsize;
  char *new_id, *c;
  struct cached_datum *newerversion;
  struct dobj *u = *uu;

  if (!cachesize)
    return;
  c = classify (uu, NULL);
  if (c)
    if (!(strcmp (c, "crap")))
      return;

  new_id = find_oid (u);
  if (!new_id)
    return;

  /* kill_old_versions will silently purge old versions of the datum in the
   * message from cache (returning NULL) or return a pointer if a newer
   * version of the datum in the message is in the cache.
   */
  if (newerversion = kill_old_versions (new_id))
  {
#ifdef debugcache
    printf ("Replacing %s with newer version found in cache\n", new_id);
    fflush (stdout);
#endif
    kill_message_datum (uu, newerversion);
    /* Clear counters too */
    c_hops = c_links = c_cache = 0;
    return;
  }

  dsize = size_dobj (u);
  if (cachesize < dsize + sizeof (struct cached_datum))
  {
#ifdef debugcache
    puts ("Cache smaller than datum!");
#endif
    /* Still need to purge old versions which could be smaller */
    return;
  }

  if (decision = decide (u, inlag, dsize, new_id))
  {
    struct cached_datum *t;
    t = (struct cached_datum *) malloc (sizeof (struct cached_datum));
    if (!t)
      panic ("Can't malloc in cache_hook", 1);
    t->timestamp = time((time_t *)0);
    t->unrealtimecached = unrealtime;
    if (decision == 2)
    {
#ifdef debugcache
      puts ("Adding warm hint to unrealtimecached");
#endif
      t->hint = warm_hint;
    }
    else
      t->hint = cold_hint;
    t->unrealtimecached += t->hint;
    t->data_type = strclone (classify (&u, NULL));
    t->datum = filter_clone_dobj (u, t->timestamp);
    if (!(t->ID = find_oid (t->datum)))
    {
      timlog ("ID-less datum rejected in cache_hook");
      if (t->data_type)
        free (t->data_type);
      free_dobj (t->datum);
      free (t);
      return;
    }
    t->datum_size = size_dobj (t->datum);
    t->descript = strclone (describe (u));
    t->total_size = total_size (t);
    if (cachesize < t->total_size)
    {
      /* Oops.  That tiny bit of extra overhead caught us. */
      free_cached_datum (t);
#ifdef debugcache
      puts ("Caught by overhead.  Not cached.");
#endif
      return;
    }
    while (freespace < t->total_size)
      make_room ();
    add_item (t);
    c_hops = c_links = c_cache = 0;
#ifdef debugcache
    puts ("Item cached; counters cleared.");
#endif
  }
#ifdef debugcache
  else
    puts ("Decided not to cache it.");
#endif
}

/* This is almost like the query interface to the other resources.
 */
void *
res_cache (char *query, char *class)
{
  if (query[0] == '(')
  {
    char *oids = NULL, *desc;
    long length=0, memlength=0;
    struct cached_datum *c;

    /* Need to make a nice neat desc out of query */
    desc = strclone (query);
    nopunct (desc);
    if (strspn (desc, " ") == strlen (desc))
    {
      free (desc);   /* If there's nothing in there, get rid of it. */
      desc = NULL;
    }

    append_to_queue (&oids, &length, "MATCHES ", 8, &memlength);

    c = data_cache;
    while (c)
    {
      if (!(struncump (c->data_type, class)))
      {
        int tmatches, tlength;
        if (!strcmp (class, "crap"))
          panic ("Assertion in res_cache", 1);
        if (desc && c->descript)
        {
          tmatches = wrdmatches (c->descript, desc, &tlength);
          if (tmatches == tlength)
          {
            append_to_queue (&oids, &length, "[", 1, &memlength);
            append_to_queue (&oids, &length, c->ID, strlen(c->ID), &memlength);
            append_to_queue (&oids, &length, "]", 1, &memlength);
          }
        }
        else if (!desc)
        {
          append_to_queue (&oids, &length, "[", 1, &memlength);
          append_to_queue (&oids, &length, c->ID, strlen(c->ID), &memlength);
          append_to_queue (&oids, &length, "]", 1, &memlength);
        }
      }
      c = c->next;
    }
    if (desc)
      free (desc);

    append_to_queue (&oids, &length, "\0", 1, &memlength);
    return (void *)oids;
  }
  else if (strncmp (query, "FETCH [", 7) == 0)
  {
    struct cached_datum *t;
    char tempoid[namlen];
    strcpy (tempoid, query+7);
    tempoid[strlen(tempoid)-1] = '\0';
    t = data_cache;
    while (t)
    {
      if (!(strcmp (t->ID, tempoid)))
      {
        hit_hint (t); /* Upgrade cold hints */
        return (void *) clone_dobj (t->datum);
      }
      t = t->next;
    }
    return (void *) strclone ("OOPS");
  }
  else
    panic ("res_cache:  What is this garbage?", 1);
}

/* This used to be cache_answer; now it just handles upgrading of hints.
 */
void
hit_hint (struct cached_datum *hit)
{
  struct cached_datum *p;

/*
  change_class (&(t->datum), hit->data_type);
  t->enddatum->next = clone_dobj (hit->datum);
  rebound (t, NULL);
*/

  /* Upgrade cold hints upon access */
  if (hit->hint == warm_hint)
    return;

#ifdef debugcache
  timlog ("Upgrading datum to warm_hint");
#endif

  /* The datum has to be re-inserted into the list to keep the list in order.
   */
  p = data_cache;
  if (p == hit)
  {
    data_cache = hit->next;
    if (!data_cache)
      tail = NULL;
  }
  else
  {
    while (p->next != hit)
    {
      p = p->next;
      if (!p)
        panic ("Assertion failure in hit_hint", 1);
    }
    p->next = hit->next;
    if (!(p->next))
      tail = p;
  }
  hit->next = NULL;
  hit->unrealtimecached = hit->unrealtimecached - hit->hint + warm_hint;
  hit->hint = warm_hint;
  freespace += hit->total_size;
  add_item (hit);
}
