/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 *  classify.c
 *
 *  Try to determine what sort of data a query is looking for.
 *
 *  Always remember that "U" is the universal class and must be treated as
 *  a special case if you wish to use it.  Using "" (null string) to designate
 *  the universal class would have been more consistent, but this would have
 *  caused other problems too numerous to mention.
 */

#include "alibi.h"

/* This function can be used by classify or by specialist resources to assign
 * a class to a classless datum.  The "crap" class should be applied to any
 * response which should not be cached and cannot help route future queries,
 * like failure messages.
 */
void
add_class (struct dobj **t, char *class)
{
  char classline[namlen+15];
  struct dobj *v;
  if (!t)
    return;
  sprintf (classline, "0 Class: %s", class);
  v = (struct dobj *) malloc (sizeof (struct dobj));
  if (!v)
    panic ("Couldn't malloc in add_class", 1);
  v->next = *t;
  v->text = strclone (classline);
  v->binlen = 0l;
  v->bindata = (char *)0;
  *t = v;
}

/* Like add_class, but for descriptions. */
void
add_descript (struct dobj **t, char *class)
{
  char classline[10000];
  struct dobj *v;
  sprintf (classline, "0 Description: %s", class);
  v = (struct dobj *) malloc (sizeof (struct dobj));
  if (!v)
    panic ("Couldn't malloc in add_descript", 1);
  v->next = *t;
  v->text = strclone (classline);
  v->binlen = 0l;
  v->bindata = (char *)0;
  *t = v;
}

/* The rabid version of add_class which destroys any existing class. */
void
change_class (struct dobj **t, char *class)
{
  struct dobj *v;
  static char classline [namlen+15];
  v = *t;
  while (v)
  {
    int flag;
    if (sscanf (v->text, "%d Class: %s", &flag, classline) == 2)
    {
      free (v->text);
      sprintf (classline, "0 Class: %s", class);
      v->text = strclone (classline);
      return;
    }
    v = v->next;
  }
  add_class (t, class);
}

/* Try to classify a query based on our query routing table.  More common
 * terms always wins.  Ties are broken by using the "shortest" (hopefully
 * most general) class.
 */
int
quclass (char *q, char *class)
{
  struct qrtable *t = qroutes;
  char *bestclass=NULL;
  int bcl_length, bcl_degree=0, t_degree, t_length;
  while (t)
  {
    t_degree = wrdmatches (q, t->data_type, &t_length);
    if (t_degree > bcl_degree || (t_degree == bcl_degree &&
    t_length < bcl_length))
    {
      bcl_degree = t_degree;
      bcl_length = t_length;
      bestclass = t->data_type;
    }
    t = t->next;
  }
  if (bcl_degree)
  {
#ifdef debugquery
    printf ("Accepted class %s with relevance %d\n", bestclass,
    bcl_degree);
    fflush (stdout);
#endif
    strcpy (class, bestclass);
    return 1;
  }
  return 0;
}

/* u should be the query text for a query and NULL for a response because
 * the response should have been given a Class line by the query-answerer.
 */
char *
classify (struct dobj **t, struct dobj *u)
{
  struct dobj *v;
  static char class [namlen];

  /* Has it already been classified? */
  if (t)
  {
    v = *t;
    while (v)
    {
      int flag;
      if (sscanf (v->text, "%d Class: %s", &flag, class) == 2)
      {
        char *fs;
        fs = strchr (v->text, ':');
        fs += 2;
        return fs;
      }
      v = v->next;
    }
  }

  /* Try to assign a classification */
  class[0] = '\0';
  if (u)
  {
    char *q;
    q = nopunct (strclone (startoftext (u)));
    /* See if it matches anything in our query routing table */
    if (quclass (q, class))
      ;
    /* Otherwise, look for known buzzwords. */
    else if (strword (q, "netnews"))
      strcpy (class, "netnews");
    else if (!(strncmp (q, "ping ", 5)))
      strcpy (class, "crap");
    free (q);
  }

  /* U = Universal Class (i.e. we're clueless) */
  if (class[0] == '\0')
    strcpy (class, "U");

  add_class (t, class);
  return class;
}

/* Descriptions are never added on-the-fly. */
char *
describe (struct dobj *t)
{
  struct dobj *v;
  static char class [namlen];

  v = t;
  while (v)
  {
    int flag;
    if (sscanf (v->text, "%d Description: %s", &flag, class) == 2)
    {
      char *fs;
      fs = strchr (v->text, ':');
      fs += 2;
      return fs;
    }
    v = v->next;
  }

  return NULL;
}
