/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 *  rblobs_gsfc.c  Customized rblobs for NASA GSFC.
 *
 *  If a file is starred in the index, it gets an IBM copyright message.  The
 *  asterisk must be first in the line, just like #.
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <stdlib.h>
#include <time.h>

void
reset_blobs ();
void
fetch ();
void
print ();
void
do_query ();
void
ditch ();
char *
fname2class ();
void
declare_all_classes ();

/* Approx. number of OID's returned for a () query */
#define universe_size 30

/* This directory only matters for the named pipes. */
#define defdir "/users/davidf/bin/dunloggin.gsfc.nasa.gov"

int infd, outfd;
long myaddr=0, idxsize=0;
double randmax;  /* RAND_MAX + 1 */

/* The following is the class of data which is prepended to the terms in the
 * directory path to form the classification of each blob.
 */
#define blob_class "blob image NASA"

/* The root blob directory. */
#define blobdir "/users/davidf/blobs"

struct bdex {
  char *file;
  char *descript;
  int ibmflag;
  struct bdex *next;
};

struct bdex *vindex = NULL;

void
shutdown ()
{
  print ("BYE\r\n");
  /* Let Unetd read the BYE */
  sleep (10);
  exit (-1);
}

void
timlog (char *a)
{
  print ("CONLOG ");
  print (a);
  print ("\r\n");
}

void
panic (char *why, int a)
{
  timlog (why);
  shutdown ();
}

#include "strfuns.c"

void
create_pipe (int *fd, char *name)
{
  *fd = -1;
  if (mknod (name, S_IFIFO | S_IREAD | S_IWRITE, NULL))
  {
    perror ("mknod");
    exit (-1);
  }
  *fd = open (name, O_RDWR, NULL);
  if (*fd == -1)
  {
    perror ("open");
    exit (-1);
  }
}

void
print (char *m)
{
  write (outfd, m, strlen (m));
}

/* Get one character from the server. */
char
sgetc ()
{
  char t;
  int l=0;
  while (!l)
  {
    l = read (infd, &t, 1);
    if (l <= -1 && errno == EINTR)
    {
      l = 0;
      continue;
    }
    if (l <= -1)
      panic ("Error on read from server -- shutting down.", 1);
  }
  return t;
}

/* Get a line of text */
char *
get_text ()
{
  static char buf[10000];
  char c='a';
  int l=0;
  while (c != '\n')
  {
    if (l >= 10000)
      panic ("Input buffer overflow.  Shutting down.", 1);
    c = sgetc ();
    buf[l++] = c;
  }
  buf[l] = '\0';
  return buf;
}

void
handler (char *m)
{
  noeol (m);
  if (strncmp (m, "RESET", 5) == 0)
  {
    if (sscanf (m, "RESET %ld", &myaddr) != 1)
    {
      timlog (m);
      timlog ("RESET didn't provide address");
    }
    reset_blobs ();
  }
  else if (m[0] == '(')
  {
    if (!myaddr)    /* Not ready yet */
    {
      timlog ("Can't answer queries until RESET provides address");
      print ("MATCHES\r\n");
    }
    else
      do_query (m);
  }
  else if (strncmp (m, "FETCH [", 7) == 0)
    fetch (m+6);
  else
  {
    print ("CONLOG What is all this garbage? (see below)\r\nCONLOG ");
    print (m);
    print ("\r\n");
  }
}

/* Inform Unetd that we are leaving. */
void
sig_shutdown (int sig)
{
  printf ("Received unexpected signal %d -- shutting down.\n", sig);
  shutdown ();
}

/* Administrator can manually send a SIGPIPE if Unetd crashes to keep from
 * having to restart the resource.
 */
void
newpipes ()
{
  signal (SIGPIPE, newpipes);
  close (infd);
  close (outfd);
  create_pipe (&infd, "wres_blob");
  create_pipe (&outfd, "rres_blob");
  print ("CONLOG new pipes\r\n");
}

void
do_sigs ()
{
  int a;
  for (a=1;a<NSIG;a++)
    if (a != SIGCONT)
      signal (a, sig_shutdown);
  signal (SIGPIPE, newpipes);
}

/******** End generic resource code, begin blob-specific stuff *******/

/* I got sick of compilers that didn't define RAND_MAX. */
void
rand_setup ()
{
  long temp,max=0,a;
  srand (time (NULL));
  for (a=0;a<100;a++)
  {
    temp = rand ();
    if (temp > max)
      max = temp;
  }
  if (max < 32768)
    randmax = 32768.0;
  else
    randmax = 2147483648.0;
}

void
declare_all_classes ()
{
  struct bdex *t;
  char prevclass[namlen];
  t = vindex;
  prevclass[0] = '\0';
  while (t)
  {
    if (strcmp (fname2class (t->file), prevclass))
    {
      strcpy (prevclass, fname2class (t->file));
      print ("DECLASS ");
      print (prevclass);
      print ("\r\n");
    }
    t = t->next;
  }
}

/* Given the full path name of a file, this converts it to the path relative
 * to the root blobdir.
 */
char *
relpath (char *fullpath)
{
  int a, b;
  a = strlen (blobdir);
  b = strlen (fullpath);
  if (strncmp (fullpath, blobdir, a))
  {
    timlog ("Already a relpath in relpath");
    return fullpath;
  }
  if (b -1 < a)
    return fullpath+b;
  if (fullpath[a] != '/')
  {
    timlog ("fullpath not relative to blobdir");
    return fullpath;
  }
  return fullpath+a+1;
}

/* Given a file name, return the classification of that file. */
char *
fname2class (char *fname)
{
  static char blam[namlen+80], *a;
  strcpy (blam, fname);
  a = strrchr (blam, '/');                  /* Remove file name from path */
  if (a)
    *a = '\0';
  a = strclone (relpath (blam));            /* Get relative path */
  if (strlen (a))
    sprintf (blam, "%s %s", blob_class, a); /* Make classification */
  else
    strcpy (blam, blob_class);
  free (a);
  a = blam;
  while (*a)
  {
    if (*a == '/')
      *a = ' ';
    a++;
  }
  return blam;
}

void
wipe_index ()
{
  while (vindex)
  {
    struct bdex *t;
    if (vindex->descript)
      free (vindex->descript);
    if (vindex->file)
      free (vindex->file);
    t = vindex->next;
    free (vindex);
    vindex = t;
  }
  idxsize = 0;
}

struct bdex *
srch_by_fname (char *fname)
{
  struct bdex *t = vindex;
  while (t)
  {
    if (!(strcmp (fname, t->file)))
      return t;
    t = t->next;
  }
  return NULL;
}

void
get_one_index_file (char *dirname, char *fname)
{
  FILE *fp;
  char bigbuf[10000], smallbuf[namlen];
  sprintf (smallbuf, "%s/%s", dirname, fname);
  if (!(fp = fopen (smallbuf, "r")))
  {
    sprintf (bigbuf, "Could not open index file %s", smallbuf);
    timlog (bigbuf);
    return;
  }
#ifdef debugquery
  printf ("Reading index file %s\n", smallbuf);
#endif
  while (fgets (bigbuf, 10000, fp))
  {
    struct bdex *t;
    char fname[namlen];
    int starflag;
    if (bigbuf[0] == '#')
      continue;
    if (bigbuf[0] == '*')
    {
      starflag = 1;
      bigbuf[0] = ' ';
    }
    else
      starflag = 0;
    noeol (bigbuf);
    sscanf (bigbuf, "%s", fname);
    sprintf (smallbuf, "%s/%s", dirname, fname);
    t = srch_by_fname (smallbuf);
    if (t)
    {
      /* Just have to add new description text */
      char *old, *new;
      old = t->descript;
      new = strchr (bigbuf, ' ');
      if (new)
      {
        t->descript = (char *) safmalloc (strlen (old) + strlen (new) + 1);
        strcpy (t->descript, old);
        strcat (t->descript, new);
        nopunct (t->descript);
        free (old);
      }
    }
    else
    {
      t = (struct bdex *) safmalloc (sizeof (struct bdex));
      t->file = strclone (smallbuf);
      strcat (bigbuf, " ");
      strcat (bigbuf, relpath (dirname));
      t->descript = strclone (bigbuf);
      t->ibmflag = starflag;
      t->next = vindex;
      vindex = t;
      nopunct (t->descript);
      idxsize++;
    }
  }
  fclose (fp);
}

void
get_index (char *dir)
{
  DIR *dirp;
  struct dirent *dp;
  struct stat s;
  char fnam[namlen];
  if (!(dirp = opendir(dir)))
  {
    char tbuf[namlen+80];
    sprintf (tbuf, "Could not open directory %s", dir);
    timlog (tbuf);
    return;
  }
  for (dp = readdir(dirp); dp != NULL; dp = readdir(dirp))
  {
    if (!strcmp (dp->d_name, ".") || !strcmp (dp->d_name, ".."))
      continue;
    sprintf (fnam, "%s/%s", dir, dp->d_name);
    if (stat (fnam, &s) != -1)
    {
      /* Read indices */
      if (S_ISREG (s.st_mode) && (!(S_ISDIR (s.st_mode))))
      {
        if (strlen (dp->d_name) > 10)
          if (!(strcmp (dp->d_name+strlen(dp->d_name)-10, "-index.txt")))
            get_one_index_file (dir, dp->d_name);
      }
      /* Recurse directories */
      else if (S_ISDIR (s.st_mode))
        get_index (fnam);
    }
    else
    {
      char bogus[namlen+80];
      sprintf (bogus, "Couldn't stat %s", fnam);
      timlog (bogus);
    }
  }
  closedir(dirp);
}

void
reset_blobs ()
{
  print ("CONLOG Blobs resource reset\r\n");
  wipe_index ();
  get_index (blobdir);
  print ("READY\r\n");
#ifdef blob_class
  print ("MYTYPE ");
  print (blob_class);
  print ("\r\n");
#endif
  declare_all_classes ();
}

void
send_fetch_oid (struct bdex *c)
{
  char temp[namlen*2];
  struct stat s;
  if (stat (c->file, &s) == -1)
  {
    /* Don't CONLOG in the middle of a FETCH!!! */
    ditch (c->file);
    return;
  }
  sprintf (temp, "[%s %ld %lu]", c->file, myaddr, (unsigned long) s.st_mtime);
  print (temp);
}

void
do_query (char *query)
{
  struct bdex *c;
  nopunct (query);
  if (strspn (query, " ") == strlen (query))
    query[0] = '\0';
  print ("MATCHES ");

  c = vindex;
  while (c)
  {
    int tmatches, tlength;
    if (strlen (query))
    {
      if (c->descript)
      {
        tmatches = wrdmatches (c->descript, query, &tlength);
        if (tmatches == tlength)
          send_fetch_oid (c);
      }
    }
    else
    {
      /* Universal query -- limit size of response */
      if ((double)rand () / randmax < (double)universe_size/(double)idxsize)
        send_fetch_oid (c);
    }
    c = c->next;
  }
  print ("\r\n");
}

/* Remove an index entry for a non-existent or problematic file. */
void
ditch (char *f2ditch)
{
  struct bdex *p, *q;
  if (!(strcmp (vindex->file, f2ditch)))
  {
    p = vindex;
    vindex = p->next;
    free (p->file);
    if (p->descript)
      free (p->descript);
    free (p);
    idxsize--;
    return;
  }
  p = vindex;
  q = p->next;
  while (q)
  {
    if (!(strcmp (q->file, f2ditch)))
    {
      p->next = q->next;
      free (q->file);
      if (q->descript)
        free (q->descript);
      free (q);
      idxsize--;
      return;
    }
    p = q;
    q = q->next;
  }
}

/* Respond to a FETCH from Unetd.
 */
void
fetch (char *oid)
{
  struct bdex *hit;
  struct stat s;
  char fname[namlen], newoid[namlen*2];
  long looper;
  FILE *fp;
  sscanf (oid, "[%s", fname);
  hit = srch_by_fname (fname);
  if (!hit)
  {
    sprintf (newoid, "Can't find '%s'", fname);
    timlog (newoid);
    print ("OOPS\r\n");
    return;
  }
  if (stat (fname, &s) == -1)
  {
    sprintf (newoid, "Can't stat '%s'", fname);
    timlog (newoid);
    print ("OOPS\r\n");
    ditch (fname);
    return;
  }
  fp = fopen (fname, "rb");
  if (!fp)
  {
    sprintf (newoid, "Can't open '%s'", fname);
    timlog (newoid);
    print ("OOPS\r\n");
    ditch (fname);
    return;
  }
  print ("0 The following datum was provided by a blobs resource.");
  if (hit->ibmflag)
  {
    print ("\r\n0 This image is copyrighted by IBM!");
  }
  print ("\r\n0 ID: ");
  sprintf (newoid, "%s %ld %lu", fname, myaddr, (unsigned long) s.st_mtime);
  print (newoid);
  print ("\r\n0 Class: ");
  print (fname2class (fname));
  print ("\r\n0 Description: ");
  print (hit->descript);
  print ("\r\n");
  sprintf (newoid, "3 Binary data follows ... %ld bytes, filename %s\r\n",
    s.st_size, strrchr(fname,'/') ? strrchr(fname,'/')+1 : fname);
  print (newoid);
/*
  for (looper=0;looper<s.st_size;looper++)
  {
    unsigned char temp;
    temp = (unsigned char) fgetc (fp);
    write (outfd, &temp, 1);
  }
*/
  /* This is faster. */
  looper=0;
  while (looper<s.st_size)
  {
    int temp;
    char buf[4096];
    temp = fread (buf, 1, 4096, fp);
    looper += temp;
    /* They say never to test for an error you don't know how to handle,
     * so we'll assume that this write succeeds.
     */
    write (outfd, buf, temp);
  }
  fclose (fp);
}

main ()
{
#ifdef defdir
  if (chdir (defdir))
  {
    printf ("Cannot change directory to %s\n", defdir);
    puts ("Shutting down.");
    exit (-1);
  }
#endif
  if (fork ()) return 0;       /* Get away from parent. */
#ifndef nosetsid
  setsid ();                   /* Get away from controlling terminal. */
#endif
  do_sigs ();
  rand_setup ();
  create_pipe (&infd, "wres_blob");
  create_pipe (&outfd, "rres_blob");
  while (1)
    handler (get_text ());
}
