/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 *  netchang.c
 *
 *  Implementation of an extension to the NETCHANGE protocol described in
 *  Tajibnapis, William D.  A Correctness Proof of a Topology Information
 *  Maintenance Protocol for a Distributed Computer Network.  Communications
 *  of the ACM, Volume 20, Number 7, July 1977, pages 477-485.
 *
 */

#include "alibi.h"
int hash ();
void netchange ();
void dtabset ();
void dtabdel ();
void ncnch ();
void broadcost ();

/* Route_table contains the current best nexthops for every destination.  To
 * find an entry, hash the name of the place you want to go and then scan the
 * linked list to find the correct destination.
 */
struct route_entry * route_table [numbuckets];

struct dtab_row {
  char dest[namlen];
  struct dtab_row *next;
  struct route_entry *columns;
};

/* Distance_table contains distances for every possible nexthop for every
 * destination.  To find an entry, hash the name of the place you want to go,
 * scan the linked list to find the correct destination, then scan the columns
 * list to go through the nexthops.
 */
struct dtab_row * distance_table [numbuckets];

void
nc_init ()
{
  int a;
  for (a=0;a<numbuckets;a++)
  {
    route_table [a] = (struct route_entry *) 0;
    distance_table [a] = (struct dtab_row *) 0;
  }
}

/* Returns null pointer on don't know. */
struct route_entry *
routeto (char *whereto)
{
  struct route_entry *this = route_table [hash (whereto)];
  while (this)
  {
    if (strcmp (whereto, this->dest) == 0)
      break;
    this = this->next;
  }
  return this;
}

/* Locate row in distance table for a host, or return null */
struct route_entry *
dtab (char *whereto)
{
  struct dtab_row *this = distance_table [hash (whereto)];
  while (this)
  {
    if (strcmp (whereto, this->dest) == 0)
      break;
    this = this->next;
  }
  if (this)
  {
    if (!(this->columns))
    {
      /* No point having an empty row. */
      int a;
      struct dtab_row *p = distance_table [a = hash (whereto)];
#ifdef lognetc
      printf ("Removing %s from distance table.\n", whereto);
      fflush (stdout);
#endif
      if (this == p)
        distance_table [a] = this->next;
      else
      {
        while (p->next != this)
        {
          p = p->next;
          if ((!p) || (!(p->next)))
            panic ("Assertion failure in dtab", 1);
        }
        p->next = this->next;
      }
      free (this);
      return (struct route_entry *)0;
    }
    return this->columns;
  }
  else
    return (struct route_entry *)0;
}

void
nc_hashstat (struct conlist *t)
{
  int a;
  char buf[100];
  for (a=0;a<numbuckets;a++)
  {
    int count=0;
    struct route_entry *this = route_table[a];
    while (this)
    {
      sprintf (buf, "[%s via %s %ld] ", this->dest, this->nexthop, this->distance);
      text_sock_write (t, buf);
      count++;
      this = this->next;
    }
    sprintf (buf, "\n^^ Bucket %d size %d\n", a, count);
    text_sock_write (t, buf);
  }
  text_sock_write (t, "\nReady.\n");
}

int
hash (char *name)
{
  int retval = 0;
  while (*name)
  {
    retval += (int)(*(name++));
    retval = (retval * mult) % numbuckets;
  }
  return retval;
}

void
delroute (struct route_entry *r, char *whereto)
{
  int a;
  struct route_entry *p = route_table [a = hash (whereto)];
#ifdef lognetc
  {
    char buf[namlen+40];
    sprintf (buf, "Removing %s from route table", whereto);
    timlog (buf);
  }
#endif
  if (r == p)
    route_table [a] = r->next;
  else
  {
    while (p->next != r)
    {
      p = p->next;
      if ((!p)||(!(p->next)))
        panic ("Assertion failure in route_entry", 1);
    }
    p->next = r->next;
  }
  free (r);
}

/* Update the routing table for a particular destination.  If there is no
 * routing table entry for that destination, one is created.  If there is no
 * row in the distance table for that destination, any existing routing table
 * entry is destroyed.  If the distance exceeds maxlag, the existing entry is
 * destroyed.
 */
void
nc_recompute_route (char *whereto)
{
  struct route_entry *r = routeto (whereto), *d = dtab (whereto);
  if (!d)
  {
    if (r)
      delroute (r, whereto);
    return;
  }
  if (!r)
  {
    int a;
    r = (struct route_entry *) malloc (sizeof (struct route_entry));
#ifdef lognetc
    {
      char buf[namlen+40];
      sprintf (buf, "Adding %s to route table", whereto);
      timlog (buf);
    }
#endif
    if (!r)
      panic ("Can't malloc in nc_recompute_route", 1);
    strcpy (r->dest, whereto);
    a = hash (whereto);
    r->next = route_table [a];
    route_table [a] = r;
  }
  r->distance = d->distance;
  strcpy (r->nexthop, d->nexthop);
  d = d->next;
  while (d)
  {
    if (d->distance < r->distance)
    {
      r->distance = d->distance;
      strcpy (r->nexthop, d->nexthop);
    }
    d = d->next;
  }
  if (r->distance > maxlag)
  {
#ifdef lognetc
    printf ("Route for %s exceeds lag limit.\n", r->dest);
#endif
    delroute (r, whereto);
  }
}

/* This function sends an entire routing table to a peer.  It is
 * used when a link comes up according to the algorithm.
 *
 *  ADDITION:  Also sends DECLASS messages for known class hierarchy to
 *    reduce painful startup period at new sites.  (Somewhat wasteful for
 *    sites that are not really new, but the show must go on)
 */
void
nc_send_table (char oid[namlen])
{
  struct conlist *c;
  int a;
  struct route_entry *x;
  c = findcon (oid);
  if (!c)
    panic ("No such peer in nc_send_table", 1);
  for (a=0;a<numbuckets;a++)
  {
    x = route_table [a];
    while (x)
    {
#ifdef lognetc
      char buf[240];
      sprintf (buf, "Sending [%s %s %ld] to", myhostname, x->dest,
        x->distance);
      conlog (c, buf);
#endif
      ncnch (myhostname, x->dest, x->distance, c);
      x = x->next;
    }
  }

  /* Send class hierarchy too.  If classes are ever stored in a tree
   * structure, just send leaf nodes.
   */
  {
    struct qrtable *a = qroutes;
    while (a)
    {
      char buf[namlen+40];
      sprintf (buf, "DECLASS %s\n", a->data_type);
      text_sock_write (c, buf);
      a = a->next;
    }
  }
}

/* ADD A LOCAL LINK */
void
nc_add_column (char oid[namlen])   /* Name of peer */
{
  char buf[namlen+40];
  sprintf (buf, "Link to %s coming up", oid);
  timlog (buf);
  dtabset (oid, oid, 1000000);   /* One second. */
  /* Must call broadcost before sending route table to c so that route */
  /* table entry for c is created. */
  broadcost (oid);
  nc_send_table (oid);
}

/* Update a row in the routing table and possibly broadcast a change in cost.
 */
void
broadcost (char oid[namlen])  /* Name of destination */
{
  long old_distance;
  struct route_entry *e;
  e = routeto (oid);
  if (e)
    old_distance = e->distance;
  else
    old_distance = -1;
  nc_recompute_route (oid);
  e = routeto (oid);
  if (e)
  {
    /* If new distance is different, broadcast it. */
    if (e->distance != old_distance)
      netchange (myhostname, oid, e->distance, (char *)0);
  }
  else if (old_distance > 0)
  {
    /* Can't get there from here anymore! */
    netchange (myhostname, oid, -1, (char *)0);
  }
}

/* DELETE A LOCAL LINK */
void
nc_del_column (char oid[namlen])   /* Name of peer */
{
  struct dtab_row *c;
  int d;
  char buf[namlen+40];
  sprintf (buf, "Link to %s going down", oid);
  timlog (buf);
  /* Physically remove column bc from the distance table */
  for (d=0;d<numbuckets;d++)
  {
    c = distance_table [d];
    while (c)
    {
      char temp[namlen];
      strcpy (temp, c->dest);
      c = c->next;
      dtabdel (oid, temp);
      broadcost (temp);
    }
  }
}

/* This is a function to update the lag to a neighboring host.  It is not
 * part of the base algorithm, but part of the extension to use variable link
 * costs.
 */
void
nc_alter_lag (struct conlist *o, long delta_lag)
{
  struct dtab_row *c;
  int d;
  char buf[80];
  if (!delta_lag)
  {
#ifdef lognetc
    conlog (o, "No change in lag from/to");
#endif
    return;
  }
#ifdef lognetc
  sprintf (buf, "Altering lag by %ld from/to", delta_lag);
  conlog (o, buf);
#endif
  for (d=0;d<numbuckets;d++)
  {
    c = distance_table [d];
    while (c)
    {
      struct dtab_row *cc = c->next;
      struct route_entry *this = c->columns;
      while (this)
      {
        struct route_entry *tthis = this->next;
        if (!(strcmp (this->nexthop, o->name)))
        {
          this->distance += delta_lag;
          broadcost (this->dest);
        }
        this = tthis;
      }
      c = cc;
    }
  }
#ifdef lognetc
  conlog (o, "Done altering lag from/to");
#endif
}

/* Place a netchange message on a particular connection. */
void
ncnch (char tid[namlen], char oid[namlen], long sd, struct conlist *t)
{
  char buf[namlen*2+80];

  /* Silently filter out useless NETCHANGEs */
  if (!(strcmp (oid, t->name)))
    return;

  sprintf (buf, "NETCHANGE %s %s %ld\n", tid, oid, sd);
  text_sock_write (t, buf);
}

/* Send a netchange message.  If sendto is null, broadcast to all neighbors. */
void
netchange (char tid[namlen], char oid[namlen], long sd,
char sendto[namlen])
{
  struct conlist *t;
  char buf[240];
  if (sendto)
  {
    t = findcon (sendto);
    if (!t)
      panic ("No such neighbor in netchange", 1);
#ifdef lognetc
    sprintf (buf, "Sending [%s %s %ld] to", tid, oid, sd);
    conlog (t, buf);
#endif
    ncnch (tid, oid, sd, t);
  }
  else
  {
#ifdef lognetc
    printf ("Sending [%s %s %ld] to all neighbors.\n", tid, oid, sd);
    fflush (stdout);
#endif
    t = cons;
    while (t)
    {
      if (t->type == daemon)
        ncnch (tid, oid, sd, t);
      t = t->next;
    }
  }
}

/* Set / create and set an entry in the distance table */
/*               nexthop            destination  */
void
dtabset (char tid[namlen], char oid[namlen], long sd)
{
  struct route_entry *d, *e;
  d = e = dtab (oid);
  if (d)
  {
    while (d)
    {
      if (strcmp (d->dest, oid))
         panic ("Assertion failure in dtabset", 1);
      if (strcmp (d->nexthop, tid) == 0)
        break;
      d = d->next;
    }
    /* Add a distance table entry if one is needed. */
    if (!d)
    {
      d = (struct route_entry *) malloc (sizeof (struct route_entry));
      if (!d)
        panic ("Can't malloc in dtabset", 1);
      strcpy (d->dest, oid);
      strcpy (d->nexthop, tid);
      d->next = e->next;
      e->next = d;
    }
  }
  else
  {
    /* Entire row of distance table is missing */
    struct dtab_row *t;
    int a;
#ifdef lognetc
    printf ("Adding %s to distance table.\n", oid);
    fflush (stdout);
#endif
    t = (struct dtab_row *) malloc (sizeof (struct dtab_row));
    if (!t)
      panic ("Can't malloc in dtabset", 1);
    a = hash (oid);
    t->next = distance_table [a];
    distance_table [a] = t;
    strcpy (t->dest, oid);
    d = (struct route_entry *) malloc (sizeof (struct route_entry));
    if (!d)
      panic ("Can't malloc in dtabset", 1);
    strcpy (d->dest, oid);
    strcpy (d->nexthop, tid);
    d->next = (struct route_entry *)0;
    t->columns = d;
  }
  d->distance = sd;
}

/* Destroy an entry in the distance table -- rather than setting to infinity */
/*               nexthop           destination */
void
dtabdel (char tid[namlen], char oid[namlen])
{
  struct route_entry *a, *b;
  struct dtab_row *c;
#ifdef lognetc
  printf ("Deleting distance table entry for %s via %s\n", oid, tid);
  fflush (stdout);
#endif
  c = distance_table [hash (oid)];
  while (c)
  {
    if (strcmp (c->dest, oid) == 0)
    {
      a = c->columns;
      b = (struct route_entry *)0;
      if (a)
      {
        if (strcmp (a->nexthop, tid) == 0)
        {
          c->columns = a->next;
          free (a);
          return;
        }
        else
        {
          while (1)
          {
            if (!a)
            {
              printf ("Not found.\n");
              fflush (stdout);
              return;
            }
            else if (strcmp (a->nexthop, tid) == 0)
            {
              if (!b)
                panic ("Logic error in dtabdel", 1);
              b->next = a->next;
              free (a);
              return;
            }
            else
            {
              b = a;
              a = a->next;
            }
          }
        }
      }
    }
    c = c->next;
  }
  printf ("Not found.\n");
  fflush (stdout);
  return;
}

/* Process a received netchange. */
void
nc_netchange (struct conlist *t, char oid[namlen], long sd)
{
  netccount++;
#ifdef lognetc
  {
    char bbbuf[namlen*2+80];
    sprintf (bbbuf, "Received netchange [%s %s %ld] with lag %ld from",
      t->name, oid, sd, t->lag);
    conlog (t, bbbuf);
  }
#endif
  if (strcmp (oid, myhostname) == 0)
    panic ("Received useless NETCHANGE", 1);
  if (sd >= 0)
    dtabset (t->name, oid, sd+(t->lag));
  else
    dtabdel (t->name, oid);
  broadcost (oid);
}

/* Update lag */
void
nc_new_lag (struct conlist *t, long lag)
{
  char buf[80];
  if (lag == t->lag)
  {
    conlog (t, "Lag unchanged for");
    return;
  }
#ifdef lognetc
  sprintf (buf, "Setting lag to %ld for", lag);
  conlog (t, buf);
#endif
  nc_alter_lag (t, lag-(t->lag));
  t->lag = lag;
  refigure_neighbor_costs ();
  if (t->pingflag)
  {
    sprintf (buf, "LAG %ld\n", lag);
    text_sock_write (t, buf);
  }
}
