/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* Networking utility functions.  Some code fragments trace their origins
 *  through several generations back to LPMud by Lars Pensj|.
 */

#include "alibi.h"

static unsigned long lasttime = 0, firsttime = 0, gmrt=0, gqacc=0, gqdel=0,
  gnoncracc=0, gnoncrdel=0, gcachedel=0, gpingrqcount=0, gpingrscount=0,
  gnetccount=0, glagcount=0, gquerycount=0, gresponsecount=0,
  ganscount=0;

/* Initialize the conlist structure for new connections and add to conlist.
 * Fields which are very connection-dependent are not set here.
 */
void
init_conlist (struct conlist *temp)
{
  temp->next = cons;
  cons = temp;
  temp->inbuffer = temp->yettosend = NULL;
  temp->inbuflen = temp->inmemsize = temp->sendlen = temp->sendmemsize = 0;
  temp->blanks = 0;
  temp->sidetrack = NULL;
  temp->bindata = NULL;
  temp->binsize = 0;
  temp->snum = serialnum++;
  temp->notify = no_notify;
  temp->datum = temp->enddatum = NULL;
  temp->dmes = garbage;
  temp->lag = 1000000;
  temp->lastrecv = time((time_t *)0);
}

/* Recalculate the sum of the costs of neighboring links. */
void
refigure_neighbor_costs ()
{
  struct conlist *t;
  t = cons;
  neighborlag = 0;
  while (t)
  {
    if (t->type == daemon)
      neighborlag += t->lag;
    t = t->next;
  }
}

/* Set options for most sockets we will use.
 *   -- Non-blocking I/O
 *   -- Keep connections alive
 *   -- Send queued messages in a timely manner
 */
void
sockopt (int socket)
{
  struct protoent *tcp;
  int tmp = 1;

  /* Set non-blocking I/O. */
#if 0
  if (ioctl(socket, FIONBIO, &tmp) == -1)
    panic ("ioctl", 1);
#endif
#ifdef __hpux
  if (fcntl (socket, F_SETFL, O_NONBLOCK) == -1)
    panic ("fcntl", 1);
#else
  if (fcntl (socket, F_SETFL, O_NDELAY) == -1)
    panic ("fcntl", 1);
#endif

  /* Socket-level option to keep connections "warm" */
  tmp = 1;
  if (setsockopt (socket, SOL_SOCKET, SO_KEEPALIVE, (char *) &tmp,
  sizeof (tmp)))
    timlog ("setsockopt failed!");

  /* TCP-level option to send messages in a timely fashion */
  if (!(tcp = getprotobyname ("tcp")))
    timlog ("getprotobyname failed!");
  else
  {
    tmp = 1;
    if (setsockopt (socket, tcp->p_proto, TCP_NODELAY, (char *) &tmp,
    sizeof (tmp)))
      timlog ("setsockopt failed!");
  }
}

/* Send a ping message. */
void
ping (struct conlist *t)
{
  char buf[100];
  if (!(t->pingflag))
    panic ("Illegal ping", 1);
  sprintf (buf, "PINGRQ %lu\n", microtime ());
#ifdef lognetc
  printf ("Sending to %s: %s", t->name, buf);
  fflush (stdout);
#endif
  text_sock_write (t, buf);
}

/* Detect and repair network failures; connect to resources;
 * update link costs; accounting.
 */
void
relink (struct conlist **cons)
{
  char buf[namlen];
  FILE *fp;
  long gqbal;

  timlog ("Relink called");
  printf ("Since last relink:\n");
  printf ("  %ld PINGRQs, %ld PINGRSs, %ld NETCHANGEs, %ld LAGs\n", pingrqcount, pingrscount, netccount, lagcount);
  printf ("  %ld query messages, %ld response messages\n", querycount, responsecount);
  printf ("  I personally answered %ld queries.\n", anscount);
  printf ("  %ld queries accepted, %ld responses delivered\n", qacc, qdel);
  printf ("  %ld \"good\" queries accepted, %ld \"good\" responses delivered, %ld from cache\n", noncracc, noncrdel, cachedel);
  if (qdel)
  {
    printf ("  Mean response time in seconds:  %f\n", (float) mrt / (float) qdel);
  }
  if (lasttime)
    printf ("  Mean # of queries answered per second:  %f\n", (float) qdel /
      (float) ((unsigned long)(time(NULL))-lasttime));
  lasttime = (unsigned long)(time(NULL));

  puts ("Global stats:");
  gqacc += qacc;
  gqdel += qdel;
  gnoncracc += noncracc;
  gnoncrdel += noncrdel;
  gcachedel += cachedel;
  gqbal = gqacc - gqdel;
  gpingrqcount += pingrqcount;
  gpingrscount += pingrscount;
  gnetccount += netccount;
  glagcount += lagcount;
  gquerycount += querycount;
  gresponsecount += responsecount;
  ganscount += anscount;
  printf ("  %ld PINGRQs, %ld PINGRSs, %ld NETCHANGEs, %ld LAGs\n", gpingrqcount, gpingrscount, gnetccount, glagcount);
  printf ("  %ld query messages, %ld response messages\n", gquerycount, gresponsecount);
  printf ("  I personally answered %ld queries.\n", ganscount);
  printf ("  %ld queries accepted, %ld responses delivered\n", gqacc, gqdel);
  printf ("  %ld \"good\" queries accepted, %ld \"good\" responses delivered, %ld from cache\n", gnoncracc, gnoncrdel, gcachedel);
  printf ("  Cumulative Balance = %ld\n", gqbal);
  if (gqbal > 0)
    printf ("    (One in every %f queries was never responded to.)\n",
      (float) gqacc / (float) gqbal);
  if (gqdel)
  {
    printf ("  Percent of all queries (good + bad) answered from cache:  %f\n",
      (float) gcachedel / (float) gqdel * 100.0);
  }
  gqbal = gnoncracc - gnoncrdel;
  printf ("  Number of \"good\" queries that got apology or no response: %ld\n",
    gqbal);
  if (gqbal > 0)
    printf ("    (One in every %f good queries was not answerable.)\n",
      (float) gnoncracc / (float) gqbal);
  gmrt += mrt;
  if (gqdel)
  {
    printf ("  Global mean response time in seconds:  %f\n", (float) gmrt / (float) gqdel);
  }
  if (firsttime)
    printf ("  Global mean # of queries answered per second:  %f\n",
      (float) gqdel /
      (float) ((unsigned long)(time(NULL))-firsttime));
  else
    firsttime = lasttime;
  cachedel=noncracc=noncrdel=mrt=qacc=qdel=netccount=pingrqcount=pingrscount=lagcount=querycount=responsecount=anscount=0;
  fflush (stdout);

  connect_resources ();

  fp = fopen ("peers", "r");
  if (fp == NULL)
    timlog ("No peers file!");
  else
  {
    while (fscanf (fp, "%s", buf) == 1)
    {
      if (strcmp (myhostname, buf))
      {
        if (duplink (cons, buf))
        {
          struct conlist *tcon = findcon (buf);
          /* If the link found by duplink is near_daemon or connecting
           * and there is no complete daemon link to the same site,
           * findcon will return NULL.
           */
          if (tcon)
            if (tcon->pingflag)
              ping (tcon);
        }
        else
          mklink (buf, cons, (struct conlist *)0);
      }
    }
  }
  fclose (fp);
  timlog ("Finished relink");
}

/* Return 1 if already connected to daemon at host.  Note that this WILL be
 * triggered by a daemon link which we have established but has not yet been
 * authenticated.  There is thus no need to recheck for duplicates when a link
 * is authenticated.
 */
int
duplink (struct conlist **cons, char *host)
{
  struct conlist *t = *cons;
  while (t != NULL)
  {
    if (strcmp (t->name, host) == 0 && (t->type == daemon ||
    t->type == near_daemon || (t->type == connecting &&
    t->n_type == near_daemon)))
      return 1;
    t = t->next;
  }
  return 0;
}

static int
give_name (struct conlist *t)
{
  switch (t->addr.sa_family) {
  case AF_INET:
  {
    char *c;
    struct hostent *hp;
    struct sockaddr_in *u = (struct sockaddr_in *)(&(t->addr));
    unsigned long a = u->sin_addr.s_addr;
    c =
#ifdef sun
#ifndef __svr4__
    inet_ntoa (&a);
#else
    inet_ntoa (u->sin_addr);
#endif
#else
    inet_ntoa (u->sin_addr);
#endif
    t->number = strclone (c);
    if ((hp = gethostbyaddr ((char *)(&a), sizeof a, AF_INET)) == NULL)
    {
#ifdef sun
      timlog ("Error with gethostbyaddr!");
#else
      herror ("gethostbyaddr");
#endif
      return 0;
    }
    else
    {
      /* Compensate for name server returning shortened host names */
      if (!strchr (hp->h_name, '.'))
      {
        t->name = safmalloc (strlen (hp->h_name) +
                             strlen (strchr (myhostname, '.')) + 1);
        sprintf (t->name, "%s%s", hp->h_name, strchr (myhostname, '.'));
      }
      else
        t->name = strclone (hp->h_name);
      strdowncase (t->name);
    }
    break;
  }
  default:
    fprintf (stderr, "Connection from address family %hu not handled\n",
      t->addr.sa_family);
    fflush (stderr);
    panic ("Weird address", 1);
  }
  return 1;
}

void
powerup ()
{
  struct sockaddr_in sin;
  int tmp;

  inport_number = -1;

#ifdef norandom
  srand (myaddr);       /* Possible conversion problems */
#else
  srandom (myaddr);
#endif

  memset ((char *) &sin, '\0', sizeof sin);
  sin.sin_family = (short)AF_INET;
  sin.sin_addr.s_addr = INADDR_ANY;
  insocket = socket (AF_INET, SOCK_STREAM, 0);
  if (insocket == -1)
    panic ("socket", 1);
  tmp = 1;
  if (setsockopt (insocket, SOL_SOCKET, SO_REUSEADDR,
    (char *) &tmp, sizeof (tmp)) < 0)
    panic ("setsockopt", 1);

#if 0
  while (inport_number == -1)
  {
    inport_number = next_port_number ();
    sin.sin_port = htons (inport_number);
    if (bind (insocket, &sin, sizeof sin) == -1)
    {
      if (errno == EADDRINUSE)
      {
        char bogus[80];
        inport_number = -1;
        sprintf (bogus, "CAN'T BIND PORT %d", inport_number);
        timlog (bogus);
        continue;
      }
      panic ("bind", 1);
    }
  }
#endif

  /* We demand to get 13592. */
  inport_number = next_port_number ();
  sin.sin_port = htons (inport_number);
  if (bind (insocket, (struct sockaddr *)(&sin), sizeof sin) == -1)
  {
    perror ("bind");
    fflush (stderr);
    panic ("Unetd must get port 13592", 1);
  }

  sockopt (insocket);
  if (listen (insocket, 5) == -1)
    panic ("listen", 1);
  FD_SET (insocket, &fdset);
  nfds = max (insocket, nfds);
}

int
get_message (char **messbuf, long *size, struct conlist *t)
{
  long l, m;
  char *buff;
  buff = *messbuf;

#ifdef logpackets
  printf ("get_message called for %s\n", t->name);
  fflush (stdout);
#endif

  if (t->type != resource)
    if (time((time_t *)0) - t->lastrecv > maxidle)
    {
      conlog (t, "Timing out connection from/to");
      return -1;
    }

  /* Try to read new data. */
  unbreak_pipe ();
  l = read (t->con, buff, (*size)-1);
  if (broke_pipe ())
  {
    conlog (t, "SIGPIPE on read from");
    return -1;
  }
  if (l <= -1)
  {
    if (errno != EWOULDBLOCK && errno != EAGAIN)
    {
      perror ("read");
      fflush (stderr);
      return -1;
    }
    l = 0;
  }
  else if (!l)
  {
#ifdef logcons
    conlog (t, "EOF");
#endif
    return -1;
  }
  else
    t->lastrecv = time((time_t *)0);
  buff[l] = '\0';

#if 0
  printf ("Packet:  '%s'\n", buff);
  if (t->inbuflen)
  {
    char dbuf[10000];
    if (t->inbuflen >= 10000)
      panic ("Packet too long for logpackets", 1);
    memcpy (dbuf, t->inbuffer, t->inbuflen);
    dbuf[t->inbuflen] = '\0';
    printf ("Still in buffer:  '%s'\n", dbuf);
  }
  fflush (stdout);
#endif

  /* Append new packet to buffered data */
  if (l)
    append_to_queue (&(t->inbuffer), &(t->inbuflen), buff, l, &(t->inmemsize));

  /* Process binary data */
  if (t->binsize)
  {
    if (t->inbuflen >= t->binsize)
    {
      m = t->binsize;
      if ((t->bindata = (char *) malloc (t->binsize)) == NULL)
        panic ("bindata malloc failed", 1);
      memcpy (t->bindata, t->inbuffer, t->binsize);
      strcpy (buff, "BIN_FINISH\r\n");
    }
    else
    {
      m = 0;
      if (l)
      {
        /* Yes, we received something. */
        strcpy (buff, "BINDATA\r\n");
        return 1;
      }
      else
        /* Nothing is happening. */
        return 0;
    }
  }
  /* Try to return messages */
  else if (m = (int) memchr (t->inbuffer, '\n', t->inbuflen))
  {
    m = m - (int)(t->inbuffer) + 1;
    if (m >= *size)
    {
      char temp[80];
      *size = m+1;
      sprintf (temp, "Expanding messbuf to %ld bytes", *size);
      timlog (temp);
      *messbuf = (char *) realloc (*messbuf, *size);
      if (!(*messbuf))
        panic ("Can't realloc to expand messbuf", 1);
    }
    buff = *messbuf;
    memcpy (buff, t->inbuffer, m);
    buff[m] = '\0';
    noeol (buff);
  }

  /* Remove received data from the input buffer */
  if (m)
  {
    shorten_queue (&(t->inbuffer), &(t->inbuflen), m, &(t->inmemsize));
    return 1;
  }

  /* Sorry, no donuts today. */
  return 0;
}

/* Write raw data (no translations) onto a socket, or queue it up at least. */
void
bin_sock_write (struct conlist *t, char *data, long len)
{
  append_to_queue (&(t->yettosend), &(t->sendlen), data, len, &(t->sendmemsize));
  flush_write (t);
}

/* Translate EOLs */
void
text_sock_write (struct conlist *t, char *buff)
{
  char *buff2;
  long length, from, to;

  /* Don't send NETCHANGEs, etc., when nuked.  Bypassed by bin_sock_write. */
  if (nuke && strncmp (buff, "nuke\n", 5) && strncmp (buff, "FILLER", 6))
    return;

  length = strlen (buff);
  buff2 = (char *) safmalloc (length*2+1);

  for (from=0,to=0; buff[from] != '\0';)
  {
    if (buff[from] == '\n')
    {
      /* Avoid duplication */
      if (to == 0)
        buff2[to++] = '\r';
      else if (buff2[to-1] != '\r')
        buff2[to++] = '\r';
    }
    buff2[to++] = buff[from++];
  }
  buff2[to] = '\0';      /* Just in case */
  bin_sock_write (t, buff2, to);
  free (buff2);
  return;
}

/* Check for new connections and accept them. */
int openfiles=0;
void
accept_new_cons (struct conlist **cons)
{
  int fd;
  struct sockaddr addr;
  int length = sizeof addr;
  if ((fd = accept (insocket, &addr, &length)) == -1)
  {
    if (errno != EWOULDBLOCK && errno != EAGAIN)
    {
      if (errno == EMFILE || errno == ENFILE)
      {
        if (!openfiles)
        {
          openfiles = 1;
          perror ("accept");
          timlog ("^^^^ This error will only be reported once to avoid filling the log file... ");
        }
      }
      else
        perror ("accept");
    }
  }
  else
  {
    struct conlist *temp;
    sockopt (fd);
    temp = (struct conlist *) malloc (sizeof (struct conlist));
    if (temp == NULL)
      panic ("Unable to allocate memory", 1);
    temp->addr = addr;
    temp->number = NULL;
    if (give_name (temp))
    {
      init_conlist (temp);
      FD_SET (fd, &fdset);
      nfds = max (fd, nfds);
      temp->con = temp->out_con = fd;
      temp->type = new;
      temp->pingflag = 0;
#ifdef logcons
      conlog (temp, "Accepted connection from");
#endif
      log_promo (temp);

      /* Hack alert:  if we accept a probable daemon connection, put off */
      /* running relink to avoid mutual bozoing. */
      if (!(photogenesis ((int)(ntohs (*(u_short *)(&(addr.sa_data[0])))))))
      {
#ifdef logcons
        timlog ("Deferring relink for probable daemon connection");
#endif
        last_relink = time((time_t *)0);
      }
    }
    else
    {
      char buf[80];
      if (temp->number)
        sprintf (buf, "Nameserver failure for %s", temp->number);
      else
        sprintf (buf, "Nameserver failure and can't translate number");
      timlog (buf);
      free (temp->number);
      free (temp);
      close (fd);
    }

    /* See if there are any more waiting */
    accept_new_cons (cons);
  }
}

/* Close and destroy a connection */
void
remove_con (struct conlist **cons, struct conlist *delme)
{
  struct conlist *p, *t;
  if ((*cons)->con == delme->con)
    *cons = delme->next;
  else
  {
    p = *cons;
    t = p->next;
    while (t != NULL)
    {
      if (t->con == delme->con)
      {
        p->next = t->next;
        break;
      }
      p = t;
      t = t->next;
    }
    if (t == NULL)
      panic ("Tried to delete nonexistent connection", 1);
  }
  FD_CLR (delme->con, &fdset);
  FD_CLR (delme->con, &wrset);
  FD_CLR (delme->out_con, &wrset);
  close (delme->con);
  close (delme->out_con);
  if (delme->inbuffer)
    free (delme->inbuffer);
  if (delme->yettosend)
    free (delme->yettosend);
  if (delme->bindata)
    free (delme->bindata);
  if (delme->sidetrack)
    free (delme->sidetrack);
  if (delme->name)
    free (delme->name);
  if (delme->number)
    free (delme->number);
  free_dobj (delme->datum);
#ifdef logcons
  conlog (delme, "Deleted connection from/to");
#endif
  free (delme);
}

/* Establish a daemon-to-daemon link (assumes port 13592) */
int
mklink (char *otherhost, struct conlist **cons, struct conlist *com)
{
  struct sockaddr_in sin, sou;
  struct hostent *t;
  int tmp, s=(-1), yourport=13592;

  /* System routines that are picky about whitespace drive me nuts */
  while (*otherhost == ' ') otherhost++;
  noeol (otherhost);

  /* Make sure other host is valid before creating a socket */
  t = (struct hostent *) gethostbyname (otherhost);
  if (t == NULL)
  {
    char buf[namlen+80];
    sprintf (buf, "mklink: can't gethost %s", otherhost);
    timlog (buf);
    return 0;
  }
  memset ((char *) &sou, '\0', sizeof sou);
  memcpy ((char *) &sou.sin_addr, t->h_addr, t->h_length);
  sou.sin_family = (short)(t->h_addrtype);
  sou.sin_port = (u_short) (htons (yourport));

  memset ((char *) &sin, '\0', sizeof sin);
  sin.sin_addr.s_addr = myaddr;
  sin.sin_family = (short)AF_INET;
  s = socket (AF_INET, SOCK_STREAM, 0);
  if (s == -1)
    panic ("socket", 1);

  /* We must bind to a nonphotogenic port number to pass authentication. */
  /* Normally the system routines are allowed to choose the port for an */
  /* active socket. */
  sin.sin_port = 0;
  reset_port_series ();
  while (sin.sin_port == 0)
  {
    sin.sin_port = htons (next_port_number ());
    if (bind (s, (struct sockaddr *)(&sin), sizeof sin) == -1)
    {
      if (errno != EADDRINUSE)
      {
        char bogus[80];
        perror ("bind");
        fflush (stderr);
        sprintf (bogus, "Unexpected error binding port %d", (int)ntohs(sin.sin_port));
        timlog (bogus);
      }
      sin.sin_port = 0;
      continue;
    }
  }

#if 0
  {
    char x[namlen+20];
    sprintf (x, "Trying %s....\n",
#ifdef sun
    inet_ntoa (&(sou.sin_addr))
#else
    inet_ntoa (sou.sin_addr)
#endif
    );
    printf ("%s", x);
    fflush (stdout);
    if (com)
      text_sock_write (com, x);
  }
#endif

  sockopt (s);
  tmp = connect (s, (struct sockaddr *)(&sou), sizeof sou);
  if (tmp == -1 && errno != EINPROGRESS && errno != EWOULDBLOCK
  && errno != EAGAIN)
  {
    perror ("connect");
    fflush (stderr);
    close (s);
    return 0;
  }
  {
    struct conlist *temp;
    temp = (struct conlist *) malloc (sizeof (struct conlist));
    if (temp == NULL)
      panic ("Unable to allocate memory in mklink", 1);
    temp->con = temp->out_con = s;
    /* We attempt here to forge a meaningful addr struct */
    temp->addr.sa_family = AF_INET;
    *(u_short *)(&(temp->addr.sa_data[0])) = htons (yourport);
    memcpy (&(temp->addr.sa_data[2]), (unsigned char *)(&(sou.sin_addr)), 4);
    temp->number = NULL;
    if (give_name (temp))
    {
      init_conlist (temp);
      temp->pingflag = 1;
      if (!tmp)
      {
        temp->type = near_daemon;  /* Important -- see duplink */
        FD_SET (s, &fdset);
        conlog (temp, "Connected to");
      }
      else
      {
        temp->type = connecting;
        temp->n_type = near_daemon;
        FD_SET (s, &wrset);
        /* conlog (temp, "Connect in progress to"); */
      }
      nfds = max (s, nfds);
      log_promo (temp);
      text_sock_write (temp, "You there?\n");
      return 1;
    }
    else
    {
      char buf[80];
      if (temp->number)
        sprintf (buf, "Nameserver failure for %s", temp->number);
      else
        sprintf (buf, "Nameserver failure and can't translate number");
      timlog (buf);
      free (temp->number);
      free (temp);
      close (s);
      return 0;
    }
  }
}

void
cat_text_file (struct conlist *t, char *fname)
{
  FILE *fp;
  if (fp = fopen (fname, "rt"))
  {
    char buf[1000];
    while (fgets (buf, 1000, fp))
      text_sock_write (t, buf);
    fclose (fp);
  }
}

/* Actually send queued data (at last) */
void
flush_write (struct conlist *t)
{
  long a;

  /* Refuse to write to a socket which is marked for deletion. */
  if (t->type == delete)
  {
    /* Dump data */
    shorten_queue (&(t->yettosend), &(t->sendlen), t->sendlen, &(t->sendmemsize));
    return;
  }

  /* Don't jump the gun on connecting sockets */
  if (t->type == connecting && (!(FD_ISSET (t->out_con, &twrset))))
    return;

  unbreak_pipe ();
  a = write (t->out_con, t->yettosend, (unsigned) (t->sendlen));
  if (broke_pipe ())
  {
    if (t->type != connecting)
      conlog (t, "SIGPIPE on write to");
#if 0
    else
      conlog (t, "Failed to connect to");
#endif
    delmark (t);
    return;
  }
  if (a < 0)
  {
    if (errno != EWOULDBLOCK && errno != EAGAIN)
    {
      perror ("write");
      fflush (stderr);
      delmark (t);
      return;
    }
    return;
  }
  if (t->type == connecting)
  {
    t->type = t->n_type;
    FD_SET (t->con, &fdset);      /* Permit reading */
    conlog (t, "Connection established to");
    log_promo (t);
  }
  if (a)
    shorten_queue (&(t->yettosend), &(t->sendlen), a, &(t->sendmemsize));

  /* Remove from wrset those connections which have been flushed.  Bozo
   * and delete are special cases; they remain in wrset until zorched so
   * that they will be properly "serviced."
   */
  if (t->type != bozo && t->type != delete)
    if ((!(t->sendlen)) && FD_ISSET (t->out_con, &wrset))
      FD_CLR (t->out_con, &wrset);

  /* Add to wrset those connections which have outstanding junk. */
  if ((t->sendlen) && (!(FD_ISSET (t->out_con, &wrset))))
    FD_SET (t->out_con, &wrset);
}

/* Find a _daemon_ connection */
struct conlist *
findcon (char *dest)
{
  struct conlist *a = cons;
  while (a)
  {
    if (strcmp (a->name, dest) == 0 && a->type == daemon)
      break;
    a = a->next;
  }
  return a;
}
