/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 *  query.c
 *
 *  Handle a query, either by answering it or by forwarding it.  Most of the
 *  work is actually done by resource.c; this file contains mainly support
 *  functions to manage the query routing tables and various forms of hemming
 *  and hawing.
 */

#include "alibi.h"

void
list_qroutes (struct conlist *t)
{
  char buf[1000];
  struct qrtable *q;
  struct qrrow *r;
  q = qroutes;
  while (q)
  {
    r = q->row;
    sprintf (buf, "%s:\n", q->data_type);
    text_sock_write (t, buf);
    if (!r)
      text_sock_write (t, "  (nil)\n");
    while (r)
    {
      if (r->nexthop)
        sprintf (buf, "  %s %f\n", r->nexthop->name, r->factor);
      else
        panic ("Null nexthop in a qrrow", 1);
      text_sock_write (t, buf);
      r = r->next;
    }
    q = q->next;
  }
  text_sock_write (t, "\nReady.\n");
}

/* This function is for declaring the existence of a data class without
 * providing any routing information for that class.  It's used on startup
 * to insure recognition of classes provided by the local resources.
 */
void
declare_class (char *class)
{
  struct qrtable *whereto;
  char *classification;
  if (!(strcmp (class, "crap")))
    return;
  classification = strclone (class);
  while (classification)
  {
    char *tc;
    whereto = qroutes;
    while (whereto)
    {
      if (!(strcmp (whereto->data_type, classification)))
        break;
      whereto = whereto->next;
    }
    if (!whereto)
    {
#ifdef debugquery
      printf ("Adding (null) qroute for data type %s\n", classification);
#endif
      whereto = (struct qrtable *) safmalloc (sizeof (struct qrtable));
      whereto->data_type = strclone (classification);
      whereto->row = NULL;
      whereto->next = qroutes;
      qroutes = whereto;
    }
    if (tc = strrchr (classification, ' '))
    {
      *tc = '\0';
    }
    else if (strcmp (classification, "U"))
      strcpy (classification, "U");
    else
    {
      free (classification);
      classification = NULL;
    }
  }
}

/* Given a qrrow, boost the number for the specified host and age the rest. */
void
update_qrrow (struct qrrow **a, struct conlist *host)
{
  struct qrrow *b;
  int flag=0;
  b = *a;
  while (b)
  {
    b->factor *= agefac;
    if (host == b->nexthop)
    {
      flag++;
      b->factor += boostfac;
    }
    b = b->next;
  }
  if (flag > 1)
    panic ("Duplicate records in qrrow", 1);
  if (!flag)
  {
    /* Add new record */
    b = (struct qrrow *) safmalloc (sizeof (struct qrrow));
    b->nexthop = host;
    b->factor = boostfac;
    b->next = *a;
    *a = b;
  }
}

/* Update routing info when forwarding a response.  The number of data classes
 * may be kept down by selectively ignoring specializations.
 */
void
qroute_hook (struct conlist *t)
{
  char *classification;
  struct qrtable *whereto = NULL;

  if (t->type != daemon || (!(strcmp (t->name, myhostname))))
  {
#ifdef debugquery
    puts ("Skipping qroute_hook for non-daemon or self");
#endif
    return;
  }
  classification = strclone (classify (&(t->datum), NULL));
  if (!(strcmp (classification, "crap")))
  {
    return;
  }
  while (classification)
  {
    char *tc;

    whereto = qroutes;
    while (whereto)
    {
      if (!(strcmp (whereto->data_type, classification)))
      {
        update_qrrow (&(whereto->row), t);
        break;
      }
      whereto = whereto->next;
    }
    if (!whereto)
    {
#ifdef debugquery
      printf ("Adding qroute for data type %s\n", classification);
#endif
      whereto = (struct qrtable *) safmalloc (sizeof (struct qrtable));
      whereto->data_type = strclone (classification);
      whereto->row = NULL;
      whereto->next = qroutes;
      qroutes = whereto;
      update_qrrow (&(whereto->row), t);
    }
    if (tc = strrchr (classification, ' '))
    {
      *tc = '\0';
    }
    else if (strcmp (classification, "U"))
      strcpy (classification, "U");
    else
    {
      free (classification);
      classification = NULL;
    }
  }
  return;
}

void
forward_query (struct conlist *nexthop, struct dobj *t)
{
  if (!nexthop)
    panic ("Null nexthop in forward_query", 1);
  while (t)
  {
    char *fs;
    text_sock_write (nexthop, t->text);
    fs = strchr (t->text, (int)' ');
    if (fs)
    {
      fs++;
      if (!(strncmp ("Query path:", fs, 11)))
      {
        /* This is not symmetric with response path handling */
        text_sock_write (nexthop, " ");
        text_sock_write (nexthop, nexthop->name);
      }
    }
    text_sock_write (nexthop, "\n");
    if (t->bindata)
      bin_sock_write (nexthop, t->bindata, t->binlen);
    t = t->next;
  }
}

/* This function is really just a wrapper for rebound, which requires a
 * struct conlist.  If you don't have a struct conlist handy, call reply
 * instead.
 */
void
reply (struct dobj **t, char *answer)
{
  struct conlist dummy;
  struct dobj *u;
  if (!t)
    panic ("Null pointer passed to reply", 1);
  if (!(*t))
    panic ("Null datum passed to reply", 1);
  dummy.datum = dummy.enddatum = u = *t;
  while (u)
  {
    dummy.enddatum = u;
    u = u->next;
  }
  dummy.type = generic;
  rebound (&dummy, answer);
  *t = dummy.datum;
}

/* Respond to query, optionally with a terse message to be added.  This
 * function predates reply, which was added to resolve incompatibilities.
 */
void
rebound (struct conlist *t, char *answer)
{
  struct dobj *u, *v, *q=NULL;
  char tbuf[namlen+80];
  anscount++;
  u = v = t->datum;
  t->datum = t->enddatum = (struct dobj *)0;
  addline (t, "0 Response path:", (char *)0, 0l);
  if (answer)
  {
    sprintf (tbuf, "0 \033[1m%s\033[0m", answer);
    addline (t, tbuf, (char *)0, 0l);
  }
  while (u)
  {
    int flag, junk1;
    char tempdest[namlen];
    unsigned long tempsnum;
    if (sscanf (u->text, "%d Question from %lu at %s", &junk1, &tempsnum,
    tempdest) == 3)
    {
      sprintf (tbuf, "0 Answer for %lu at %s", tempsnum, tempdest);
      addline (t, tbuf, (char *)0, 0l);
      t->forsnum = tempsnum;
      strcpy (t->forsite, tempdest);
    }
    else
    {
      char *fs;
      if (sscanf (u->text, "%d", &flag) != 1)
        panic ("No flag in rebound", 1);
      fs = startoftext (u);
      if (flag&4)
        q = u;
      if ((fs)&&(!(flag&4)))   /* Delete the question */
      {
        int digs;
        char *z;
        flag -= (flag&1);
        if (flag < 10)
          digs = 1;
        else
          digs = (int)(log10((double)flag))+1;
        z = (char *) safmalloc (strlen (fs) +digs+2);
        sprintf (z, "%d %s", flag, fs);
        addline (t, z, u->bindata, u->binlen);
        free (z);
      }
    }
    u = u->next;
  }

  /* Last chance to classify */
  if (!classify (&(t->datum), q))
    add_class (&(t->datum), "crap");

  addline (t, "1 Have a nice day =|-)", (char *)0, 0l);
  free_dobj (v);

  /* Send it */
  initiate_response (&(t->datum), t->forsite, t->forsnum);
}

/* When all else fails, choose a victim (nearly) at random.
 */
struct conlist *
anybody (char *qpath)
{
  int numcons, allcons;
  struct conlist **conarr, *u;
  char *lastguy, prevguy[namlen];

  /* Count neighbors and those who have not yet seen the query */
  numcons = allcons = 0;
  u = cons;
  while (u)
  {
    if (u->type == daemon)
    {
      allcons++;
      if (!(strword (qpath, u->name)))
        numcons++;
    }
    u = u->next;
  }
  if (!allcons)
    return NULL;

  /* If possible, make list of all neighbors who have not yet seen the query */
  if (numcons)
  {
    int a=0;
    conarr = (struct conlist **) safmalloc (numcons * sizeof (struct conlist *));
    u = cons;
    while (u)
    {
      if (u->type == daemon)
        if (!(strword (qpath, u->name)))
          conarr[a++] = u;
      u = u->next;
    }
    if (a != numcons)
      panic ("Failed assertion 1 in anybody", 1);
  }
  /* If there are none, make list of all neighbors, excluding the */
  /* last guy to see the query if possible. */
  else
  {
    int a=0;
    numcons = allcons;
    conarr = (struct conlist **) safmalloc (numcons * sizeof (struct conlist *));
    u = cons;
    lastguy = qpath + strlen(qpath) -1;
    while (*lastguy != ' ')
      lastguy--;
    while (*(lastguy-1) != ' ')
      lastguy--;
    sscanf (lastguy, "%s", prevguy);
    while (u)
    {
      if (u->type == daemon)
      {
        if (allcons == 1 || strcmp (u->name, prevguy))
          conarr[a++] = u;
        else
          numcons--;
      }
      u = u->next;
    }
    if (a != numcons)
      panic ("Failed assertion 2 in anybody", 1);
  }

  /* Choose from list */
  u = conarr[(int)(rnd () * (float)numcons)];

  /* Free list and return */
  free (conarr);
  return u;
}

void
fixup_qrrow (struct qrrow **a)
{
  struct qrrow *p, *q;
  p = *a;
  q = NULL;
  while (p)
  {
    if (!(p->nexthop))
      panic ("Null nexthop in fixup_qrrow", 1);
    if (p->nexthop->type != daemon)
    {
      if (!q)
      {
        *a = p->next;
        free (p);
        p = *a;
      }
      else
      {
        q->next = p->next;
        free (p);
        p = q->next;
      }
    }
    else
    {
      q = p;
      p = p->next;
    }
  }
}

/* This function is called to update qroutes when it contains references to
 * sites which are no longer linked to us.
 */
void
fixup_qroutes ()
{
  struct qrtable *o = qroutes;
#ifdef debugquery
  puts ("fixup_qroutes called");
#endif
  while (o)
  {
    fixup_qrrow (&(o->row));
    o = o->next;
  }
}

char *
get_qpath (struct dobj *v)
{
  char *qpath=NULL;
  while (v)
  {
    int flag;
    char junk[namlen];
    if (sscanf (v->text, "%d Query path: %s", &flag, junk) == 2)
    {
      qpath = strchr (v->text, ':');
      qpath++;
      return qpath;
    }
    v = v->next;
  }
  panic ("No query path in get_qpath", 1);
}

struct conlist *
scanrow (struct qrrow *a, char *qpath)
{
  struct conlist *best = NULL;
  float bestfac = -1.0;
  while (a)
  {
    if (!strword (qpath, a->nexthop->name))
    {
      if ((a->factor > bestfac) ||
        ((a->factor == bestfac) && (a->nexthop->lag < best->lag)))
      {
        best = a->nexthop;
        bestfac = a->factor;
#ifdef debugquery
        printf ("%s has a factor of %f.\n", best->name, bestfac);
#endif
      }
    }
#ifdef debugquery
    else
      printf ("Already been to %s....\n", a->nexthop->name);
#endif
    a = a->next;
  }
  return best;
}

/* One layer up from forward_query */
int
forward (struct dobj **t, struct dobj *u)
{
  char *classification, *qpath;
  struct qrtable *whereto = NULL;
  struct conlist *nexthop=NULL;

  qpath = get_qpath (*t);
  classification = strclone (classify (t, u));
#ifdef debugquery
  if (classification)
    printf ("Query classified as %s\n", classification);
  else
    printf ("Query not classified!\n");
#endif
  while (classification)
  {
    char *tc;
    whereto = qroutes;
    while (whereto)
    {
      if (!(strcmp (whereto->data_type, classification)))
      {
#ifdef debugquery
        puts ("Found qroute");
#endif
        if (nexthop = scanrow (whereto->row, qpath))
        {
          free (classification);
          classification = NULL;
        }
        break;
      }
      whereto = whereto->next;
    }

    /* If the first try fails, we generalize the class (remove one
     * word from the end of the classification) and try again.
     */
    if (classification)
    {
      if (tc = strrchr (classification, ' '))
      {
        *tc = '\0';
      }
      else if (strcmp (classification, "U"))
        strcpy (classification, "U");
      else
      {
        free (classification);
        classification = NULL;
      }
    }
  }

  if (!nexthop)
    nexthop = anybody (qpath);
  if (!nexthop)
  {
#ifdef debugquery
    puts ("Failed to forward query.");
    fflush (stdout);
#endif
    return 0;
  }

  if (nexthop->type != daemon)
    panic ("Failed assertion in forward", 1);

  /* Forward the query */
#ifdef debugquery
  printf ("Forwarding query to %s\n", nexthop->name);
  fflush (stdout);
#endif
  forward_query (nexthop, *t);
  return 1;
}

/* Detect query path lines and return 1 if they are too long.  Also sets
 * beenhere if the query has been here before.
 */
static int
qexpired (struct dobj *t, int *beenhere)
{
  char *fs;
  fs = startoftext (t);
  if (!(strncmp ("Query path:", fs, 11)))
  {
    char tnam[namlen+2];
    int a, b=0;
    *beenhere = 0;
    strcpy (tnam, " ");
    strcat (tnam, myhostname);
    strcat (tnam, " ");
    if (cstrstr (fs, tnam))
      *beenhere = 1;
    for (a=0;a<strlen(fs);a++)
      if (fs[a] == ' ')
        b++;
    if (b < 2+TTL_query)
      return 0;
    return 1;
  }
  return 0;
}

void
fail_query (struct dobj **t)
{
  change_class (t, "crap");
  reply (t, "All links are down.  Try again later.");
#ifdef debugquery
  timlog ("Failed to handle a query");
#endif
}

/* This function does a bit of hemming and hawing before the query is given
 * over to resource.c
 */
void
handle_query (struct conlist *t)
{
  struct dobj *u=NULL, *v;
  int beenhere=0;

  /* unrealtime is measured based on the number of queries processed */
  unrealtime++;

  /* Null t is used to initialize resources. */
  if (t)
  {
    querycount++;
    v = t->datum;
    if (!v)
      panic ("No datum in handle_query", 1);

    while (v)
    {
      int flag;

      /* Return queries which have run out of time to live */

      if (qexpired (v, &beenhere))
      {
        change_class (&(t->datum), "crap");
        rebound (t, "Query not answered (too many hops).  Either there's no answer, or the answer is not available right now.");
        return;
      }

      /* Find the line that has the query */
      /* The code: (bits)
       *   1 = last item in message
       *   2 = start of data block
       *   4 = question
       */

      if (sscanf (v->text, "%d", &flag) == 1)
        if (flag & 4)
          u=v;
      v = v->next;
    }

    if (!u)
    {
      timlog ("No question in datum in handle_query");
      rebound (t, "Say WHAT?");
      return;
    }

    /* Attempt to reclassify data with overly general classification */

    {
      char *class;
      class = classify (&(t->datum), u);
      if (!(strcmp (class, "U")))
      {
        char *newclass;
        newclass = strclone (classify (NULL, u));
        if (newclass)
        {
          if (strcmp (newclass, class))
          {
#ifdef debugquery
            printf ("Changing class %s to class %s\n", class, newclass);
#endif
            change_class (&(t->datum), newclass);
          }
          free (newclass);
        }
      }
      else
        /* Enhance the propagation of the class hierarchy */
        declare_class (class);
    }
  }

  /* Hemming and hawing complete.  Get on with it. */
  resource_query_inbox (t, u, beenhere);
}
