/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 *  resource.c   Resource wrangler.
 *
 *  Resources make themselves available by creating two named pipes in the
 *  working directory called rx and wx where x is the name of the resource.
 *  For example, rcache is for the main program to read data from the cache
 *  resource, and wcache is for the main program to write data to the cache
 *  resource.  It is suggested that both pipes be opened in read-write mode
 *  to avoid blocking or bombing out on the two opens, regardless of the
 *  state of O_NONBLOCK.
 *
 *  Commands the resource must parse:
 *     RESET %ld       (provides ID number, activates resource)
 *     FETCH [OID]
 *     (implied conjunctive keywords query)
 *
 *  The degenerate query () should be replied to with a whole bunch of random
 *  OID's, not necessarily all of them.  Except where explicitly specified by
 *  the (stupid) user, it is used only to convert queries that reduce to a
 *  negation, i.e. "give me anything but foo."
 */

#include "alibi.h"

void
free_boolstack ();
void
free_opqueue ();
void
free_res ();
void
grind_booleans ();
void
nextres ();
void
q_answer ();
void
grind_query ();

/* Some additional scaffolding is required to keep track of operations lined
 * up for resources.  The alternative is to add yet more junk to conlist that
 * is useless most of the time.
 */

struct boolstack {
  char *item;
  int negate;
  struct boolstack *next;
};

struct opqueue {
  struct dobj *datum;
  struct dobj *query;
  struct boolstack *stack;
  struct boolstack *exp;
  struct opqueue *next;
};

struct resqueue {
  struct conlist *t;
  struct opqueue *ops;
  int ready;
  char *restype;
  struct resqueue *next;
};

struct resqueue *
find_resource ();

struct resqueue *resources = NULL;

/* This is called in a couple of places to add a resource for the cache. */
void
add_cache ()
{
  if (!resources)
  {
    if (cachesize > 0)
    {
      resources = (struct resqueue *) safmalloc (sizeof (struct resqueue));
      resources->next = NULL;
      resources->restype = strclone ("U");
      resources->t = (struct conlist *) safmalloc (sizeof (struct conlist));
      resources->t->datum = resources->t->enddatum = NULL;
      resources->ready = 1;
      resources->t->name = strclone ("local_cache");
      resources->ops = NULL;
    }
  }
}

void
ill_formed_query (struct opqueue *o)
{
  change_class (&(o->datum), "crap");
  reply (&(o->datum), "Ill formed query, try HELP");
  free_opqueue (o);
}

void
add_exp (struct opqueue *o, struct boolstack **last, char *c)
{
  struct boolstack *b;
  b = (struct boolstack *) safmalloc (sizeof (struct boolstack));
  b->next = NULL;
  b->negate = 0;
  b->item = strclone (c);
  if (o->exp)
    (*last)->next = b;
  else
    o->exp = b;
  *last = b;
}

/* Parse query, create expstack, start it going. */
void
start_query (struct resqueue *u, struct opqueue *o)
{
  char *ptr;
  struct boolstack *last = NULL;
#ifdef debugbool
  printf ("Starting query for resource %s\n", u->t->name);
#endif
  free_boolstack (o->stack);
  free_boolstack (o->exp);
  o->stack = NULL;
  o->exp = NULL;
  o->next = NULL;
  ptr = startoftext (o->query);
  while (*ptr)
  {
    if (*ptr == '&' || *ptr == '|' || *ptr == '~')
    {
      char temp[2];
      temp[0] = *ptr;
      temp[1] = '\0';
      add_exp (o, &last, temp);
      ptr++;
    }
    else if (*ptr == '(' || *ptr == '[')
    {
      char *p2, temp;
      if (*ptr == '(')
        p2 = strchr (ptr+1, ')');
      else
        p2 = strchr (ptr+1, ']');
      if (!p2)
      {
        ill_formed_query (o);
        return;
      }
      p2++;
      temp = *p2;
      *p2 = '\0';
      add_exp (o, &last, ptr);
      *p2 = temp;
      ptr = p2;
    }
    else
    {
      ill_formed_query (o);
      return;
    }
  }
  grind_booleans (u, o);
}

/* Move to the back of the line. */
void
add_op (struct resqueue *u, struct opqueue *o)
{
  struct opqueue *p;
  o->next = NULL;
  p = u->ops;
  if (p)
  {
    if (p == o)
      panic ("Failed assertion in add_op", 1);
    while (p->next)
      p = p->next;
    p->next = o;
  }
  else
    u->ops = o;
}

void
get_rid_of_query (struct conlist *t)
{
  struct resqueue *u;
  struct opqueue *o;
  u = find_resource (t);
  if (!u)
    panic ("Non-existent resource in get_rid_of_query", 1);
  o = u->ops;
  if (o)
  {
    if (!(forward (&(o->datum), NULL)))
      fail_query (&(o->datum));
    u->ops = u->ops->next;
    o->next = NULL;
    free_opqueue (o);
  }
  else
    panic ("Told to get rid of query with nothing on opqueue", 1);
}

/* The "ping" resource */
int
res_ping (struct conlist *t, struct dobj *u)
{
  char whereto[namlen+10], *fs;

  if (!t)
    return 0;

  fs = startoftext (u);
  if (sscanf (fs, "ping %s", whereto) != 1)
    return 0;
  if (strcmp (whereto, myhostname))
  {
    struct route_entry *rent;
    if (!(rent = routeto (whereto)))
    {
      rebound (t, "Can't get there from here.");
      return 1;
    }
    forward_query (findcon (rent->nexthop), t->datum);
  }
  else
  {
    char buf[80];
    time_t ttime = time(NULL);
    sprintf (buf, "I'm awake.  Local time is %.24s.", ctime(&ttime));
    change_class (&(t->datum), "crap");
    rebound (t, buf);
  }
  return 1;
}

/* If none of the specialized resources handled the query, this "resource"
 * tries to forward it intelligently.
 */
int
res_forward (struct conlist *t, struct dobj *u)
{
  if (!t)
    return 0;
  return forward (&(t->datum), u);
}

/* Given a connection, find the resource struct that corresponds to it. */
struct resqueue *
find_resource (struct conlist *t)
{
  struct resqueue *u;
  u = resources;
  while (u)
  {
    if (u->t == t)
      break;
    u = u->next;
  }
  return u;
}

void
free_boolstack (struct boolstack *t)
{
  struct boolstack *u;
  while (t)
  {
    u = t;
    t = t->next;
    if (u->item)
      free (u->item);
    free (u);
  }
}

void
free_opqueue (struct opqueue *ops)
{
  struct opqueue *t;
  while (ops)
  {
    t = ops;
    ops = ops->next;
    free_dobj (t->datum);
    free_boolstack (t->stack);
    free_boolstack (t->exp);
    free (t);
  }
}

void
free_res (struct resqueue *t)
{
  if (!t)
    return;
  if (t->restype)
    free (t->restype);
  free_opqueue (t->ops);
  free (t);
}

/* Add a resource */
void
resource_up_notify (struct conlist *t)
{
  struct resqueue *u;
  add_cache ();
  u = resources;
  while (u)
  {
    if (u->t == t)      /* Verify that it is not already in there */
      return;
    u = u->next;
  }
  u = (struct resqueue *) safmalloc (sizeof (struct resqueue));
  u->ops = NULL;
  u->restype = strclone ("U");
  u->ready = 0;
  u->t = t;

  /* Keep the cache at the front of the queue for performance */
  if (resources)
  {
    if (!(strcmp ("local_cache", resources->t->name)))
    {
      u->next = resources->next;
      resources->next = u;
    }
    else
    {
      u->next = resources;
      resources = u;
    }
  }
  else
  {
    u->next = resources;
    resources = u;
  }

  conlog (t, "Resource coming up");
  {
    char temp[256];
    sprintf (temp, "RESET %ld\n", myaddr);
    text_sock_write (t, temp);
  }
}

/* Delete a resource */
void
resource_down_notify (struct conlist *t)
{
  struct resqueue *u, *p;
  conlog (t, "Resource going down");
  u = resources;
  p = NULL;
  while (u)
  {
    if (u->t == t)
    {
      /* Recover any queries that were left hanging */
      while (u->ops)
      {
        struct opqueue *o;
        o = u->ops;
        u->ops = u->ops->next;
        nextres (u, o);
      }
      /* Delete it */
      if (p)
        p->next = u->next;
      else
        resources = u->next;
      free_res (u);
      return;
    }
    p = u;
    u = u->next;
  }
}

/* Connect with a specific resource.  Two named pipes must be opened for each
 * resource.  They are unlinked as soon as they are opened.
 */
void
connect_resource (char *fname)
{
  char inname[namlen], outname[namlen];
  int infd= -1, outfd= -1;
  if (!(fname[0] == 'r' || fname[0] == 'w'))
  {
    char buf[namlen+80];
    sprintf (buf, "Illegal named pipe:  %s\n", fname);
    timlog (buf);
    /* Could unlink it and continue, but let's be cautious */
    panic ("Illegal named pipe", 1);
  }
  strcpy (inname, fname);
  strcpy (outname, fname);
  inname[0] = 'r';
  outname[0] = 'w';
  infd = open (inname, O_RDWR | O_NONBLOCK, NULL);
  outfd = open (outname, O_RDWR | O_NONBLOCK, NULL);
  if (infd != -1 && outfd != -1)
  {
    struct conlist *temp;
    temp = (struct conlist *) safmalloc (sizeof (struct conlist));
    temp->con = infd;
    temp->out_con = outfd;
    temp->name = strclone (inname+1);
    temp->number = NULL;
    temp->type = resource;
    init_conlist (temp);
    temp->pingflag = 0;
    temp->notify = resource_down_notify;
    FD_SET (infd, &fdset);
    nfds = max (infd, max (outfd, nfds));
    log_promo (temp);
    resource_up_notify (temp);
  }
  else
  {
    perror ("open");
    fflush (stderr);
    if (infd != -1)
      close (infd);
    if (outfd != -1)
      close (outfd);
  }

  /* Unlink the named pipes so they will not clutter the file system. */
  if (unlink (inname) || unlink (outname))
  {
    perror ("unlink");
    fflush (stderr);
    /* This is a fatal error because there is no way for us to know the
     * next time around that we have already opened this resource.
     */
    panic ("Couldn't unlink named pipe (resource)", 1);
  }
}

/* Connect with any waiting resources. */
void
connect_resources ()
{
  DIR *dirp;
#ifdef nodirent
  struct direct *dp;
#else
  struct dirent *dp;
#endif
  struct stat s;
  int flag=1;

  /* Resources appear as named pipes in our default directory.  The extra
   * loop is there because connect_resource can unlink pipes that remain
   * to be read in our directory structure.
   */

  while (flag)
  {
    flag = 0;
    if (!(dirp = opendir(".")))
    {
      timlog ("ERROR:  Couldn't open working directory to look for resources");
      return;
    }
    for (dp = readdir(dirp); (dp != NULL) && (!flag); dp = readdir(dirp))
    {

      /* Curse NFS!  Deleted pipes sometimes appear as hidden files. */
      if ((dp->d_name)[0] != 'r' && (dp->d_name)[0] != 'w')
        continue;

      if (stat (dp->d_name, &s) != -1)
      {
        if (S_IFIFO & s.st_mode)
        {
          connect_resource (dp->d_name);
          flag = 1;
        }
      }
      else
      {
        char bogus[namlen+80];
        sprintf (bogus, "Couldn't stat %s", dp->d_name);
        timlog (bogus);
      }
    }
    closedir(dirp);
  }
}

/* Switch a query to the next resource in line. */
void
nextres (struct resqueue *u, struct opqueue *o)
{
  if (!u)
    u = resources;
  else
    u = u->next;
  if (u)
  {
    char *c;
    c = classify (&(o->datum), o->query);
    if ((u->ready) &&
      ((!(struncump (u->restype, c)))||(!(struncump (c, u->restype)))))
      start_query (u, o);
    else
      nextres (u, o);
  }
  else
  {
    if (!(forward (&(o->datum), NULL)))
      fail_query (&(o->datum));
    o->next = NULL;
    free_opqueue (o);
  }
}

/* The top of the boolstack now has the evaluated result.  Retrieve an
 * answer.
 */
void
get_answer (struct resqueue *u, struct opqueue *o)
{
  if (!(o->stack))
    panic ("Missing operation stack", 1);
  if (o->stack->item)
  {
    char oid2get[namlen];
    /* Choose candidate at random */
    {
      char *c,*d;
      int numoids=0;
      c = o->stack->item;
      while (*c)
      {
        if (*c == '[')
          numoids++;
        c++;
      }
      numoids = (int) (rnd() * numoids);
      c = o->stack->item;
      for (;numoids;numoids--)
      {
        c++;
        while (*c != '[')
          c++;
      }
      d = oid2get;
      while (*c != ']')
      {
        *d = *c;
        c++;
        d++;
      }
      *(d++) = ']';
      *d = '\0';
    }
    /* Send for it */
    if (strcmp (u->t->name, "local_cache"))
    {
      text_sock_write (u->t, "FETCH ");
      text_sock_write (u->t, oid2get);
      text_sock_write (u->t, "\n");
      free_boolstack (o->stack);
      o->stack = NULL;
      add_op (u, o);
    }
    else
    {
      char temp[namlen+10];
      void *result;
      /* Special case for the cache */
      sprintf (temp, "FETCH %s", oid2get);
      result = res_cache (temp, NULL);
      if (*((char *)result) == 'O')    /* OOPS */
        nextres (u, o);
      else
      {
        u->t->datum = u->t->enddatum = (struct dobj *)result;
        while (u->t->enddatum->next)
          u->t->enddatum = u->t->enddatum->next;
        free_boolstack (o->stack);
        o->stack = NULL;
        add_op (u, o);
        q_answer (u->t);
      }
    }
  }
  else
  {
    /* No data satisfy the query.  Next resource, please. */
    nextres (u, o);
  }
}

/* Send a query to a resource and expect a response */
void
eval_paren_exp (struct resqueue *u, struct opqueue *o)
{
  char *temp, *class;

  temp = strclone (o->exp->item);
  if (!(strcmp (u->t->name, "local_cache")))
  {
    void *result;
    /* Rationale:  cache gets class info separately. */
    class = classify (&(o->datum), o->query);
    if (class)
    {
      sremall (temp, class);
      /* if (strspn (temp, " ()") == strlen (temp)) */
    }
#ifdef debugquery
    printf ("res_cache temp = '%s' class = '%s'\n", temp, class);
#endif
    result = res_cache (temp, class);
    free (temp);
    add_op (u, o);
    grind_query ((char *)result, u->t);
    if (result)
      free (result);
  }
  else
  {
    /* Rationale:  rblobs declares classes for subdirectories, but without
     * those directory names, it won't find the right data.
     */
    sremall (temp, u->restype);
#ifdef debugquery
    printf ("%s query = '%s'\n", u->t->name, temp);
#endif
    text_sock_write (u->t, temp);
    text_sock_write (u->t, "\n");
    free (temp);
    add_op (u, o);
  }
}

/* Combined set intersection and subtraction routine */
char *
oid_set_arith (char *a, char *b, int subtract)
{
  char temp, *v, *c;

  /* Get the degenerate cases out of the way */
  if (!a)
    return NULL;
  if (!b)
  {
    if (subtract)
      return strclone (a);
    else
      return NULL;
  }

  v = (char *) safmalloc (strlen (a) +1);
  *v = '\0';
  while (*a)
  {
    c = strchr (a+1, ']');
    if (!c)
      panic ("Assert in oid_set_arith", 1);
    c++;
    temp = *c;
    *c = '\0';
    if (subtract)
    {
      if (!(sstrstr (b, a)))
        strcat (v, a);
    }
    else
      if (sstrstr (b, a))
        strcat (v, a);
    a = c;
    *a = temp;
  }
  if (*v == '\0')
  {
    free (v);
    v = NULL;
  }
  return v;
}

/* Set addition -- no negation!  Use oid_set_arith to subtract.
 */
char *
oid_set_merge (char *a, char *b)
{
  char temp, *v, *c, *mrg;
  if ((!a)&&(!b))
    return NULL;
  if (!a)
    return strclone (b);
  if (!b)
    return strclone (a);

  mrg = (char *) safmalloc (strlen(a)+strlen(b)+1);
  strcpy (mrg, a);
  strcat (mrg, b);

#ifdef slow_Booleans
  /* All this is just to eliminate duplicates.  Not _really_ necessary. */
  v = (char *) safmalloc (strlen (mrg) +1);
  *v = '\0';
  a = mrg;
  while (*a)
  {
    c = strchr (a+1, ']');
    if (!c)
      panic ("Assert in oid_set_merge", 1);
    c++;
    temp = *c;
    *c = '\0';
    if (!(sstrstr (v, a)))
      strcat (v, a);
    a = c;
    *a = temp;
  }
  free (mrg);
  return v;
#else
  return mrg;
#endif
}

/* Boolean logic.  Negation is handled specially to avoid "subtracting from
 * the universe" except where the query reduces to a negation (give me anything
 * except foo).
 */
void
grind_booleans (struct resqueue *u, struct opqueue *o)
{
  struct boolstack *op1, *op2, *t;
  while (1)
  {
#ifdef debugbool
    {
      op1 = o->exp;
      printf ("Expression stack:  ");
      while (op1)
      {
        if (op1->negate)
          printf ("~");
        if (op1->item)
          printf ("'%s' ", op1->item);
        else
          printf ("[] ");
        op1 = op1->next;
      }
      printf ("\n");
      fflush (stdout);
    }
    {
      op1 = o->stack;
      printf ("Boolstack:  ");
      while (op1)
      {
        if (op1->negate)
          printf ("~");
        if (op1->item)
          printf ("'%s' ", op1->item);
        else
          printf ("[] ");
        op1 = op1->next;
      }
      printf ("\n");
      fflush (stdout);
    }
#endif

    t = o->exp;
    if (!t) /* Almost done */
    {
      if (!(o->stack)) /* Error if no result on stack */
      {
        ill_formed_query (o);
        return;
      }
      if (o->stack->next) /* Error if multiple results on stack */
      {
        ill_formed_query (o);
        return;
      }
      if (o->stack->item)
      {
        if ((o->stack->item)[0] != '[')
        {
          ill_formed_query (o);
          return;
        }
      }
      if (o->stack->negate)
      {
        /* Convert queries that reduce to negations by anding with an
         * approximation of U.
         */
        if (o->stack->item)
        {
          struct boolstack *last = NULL;
          add_exp (o, &last, "()");
          add_exp (o, &last, "&");
          continue;
        }
        /* ~[] must be handled specially to avoid looping */
        o->stack->item = strclone ("()");
        o->exp = o->stack;
        o->exp->negate = 0;
        o->stack = NULL;
        continue;
      }
      /* All done */
      get_answer (u, o);
      return;
    }

    if (t->item)
    {
      if ((t->item)[0] == '(')
      {
        /* Evaluate subquery and come back later */
        eval_paren_exp (u, o);
        return;
      }
      if ((t->item)[0] == '[')
      {
        /* Push result */
        o->exp = o->exp->next;
        t->next = o->stack;
        o->stack = t;
        continue;
      }
      else if ((t->item)[0] == '~')
      {
        o->exp = o->exp->next;
        free (t->item);
        free (t);
        if (!o->stack)
        {
          ill_formed_query (o);
          return;
        }
        o->stack->negate = 1 - o->stack->negate;
        continue;
      }
      else if ((t->item)[0] == '&')
      {
        char *temp;
        o->exp = o->exp->next;
        free (t->item);
        free (t);
        if (!o->stack)
        {
          ill_formed_query (o);
          return;
        }
        if (!o->stack->next)
        {
          ill_formed_query (o);
          return;
        }
        op1 = o->stack;
        op2 = op1->next;
        if (op1->negate && op2->negate)
          temp = oid_set_merge (op1->item, op2->item);
        else if (op1->negate)
          temp = oid_set_arith (op2->item, op1->item, 1);
        else if (op2->negate)
        {
          temp = oid_set_arith (op1->item, op2->item, 1);
          op2->negate = 0;
        }
        else
          temp = oid_set_arith (op1->item, op2->item, 0);
        if (op2->item)
          free (op2->item);
        op2->item = temp;
        o->stack = op2;
        if (op1->item)
          free (op1->item);
        free (op1);
        continue;
      }
      else if ((t->item)[0] == '|')
      {
        struct boolstack *b;
        if (!o->stack)
        {
          ill_formed_query (o);
          return;
        }
        if (!o->stack->next)
        {
          ill_formed_query (o);
          return;
        }
        op1 = o->stack;
        op2 = op1->next;
        /* Transform all unions into intersections :-) */
        op1->negate = 1 - op1->negate;
        op2->negate = 1 - op2->negate;
        (t->item)[0] = '~';
        b = (struct boolstack *) safmalloc (sizeof (struct boolstack));
        b->item = strclone ("&");
        b->negate = 0;
        b->next = t;
        o->exp = b;
        continue;
      }
      else
      {
        ill_formed_query (o);
        return;
      }
    }
    else
    {
      /* Push null result */
      o->exp = o->exp->next;
      t->next = o->stack;
      o->stack = t;
      continue;
    }
  }
}

/* 
/* Act on the response from the resource ("MATCHES....") */
void
grind_query (char *buf, struct conlist *t)
{
  struct resqueue *u;
  struct opqueue *o;
  u = find_resource (t);
  if (!u)
    panic ("Non-existent resource in grind_query", 1);
  o = u->ops;
  if (o)
  {
    if (!(o->exp))
      panic ("Missing expression stack", 1);
    if (!(o->exp->item))
      panic ("Missing expression", 1);
    if ((o->exp->item)[0] == '(')
    {
      /* Replace subquery with result */
      free (o->exp->item);
      if (!(strchr(buf,'[')))    /* No matches? */
        o->exp->item = NULL;
      else
        o->exp->item = strclone (tstartoftext (buf));
    }
    else
      conlog (t, "Bogus operation stack");
    /* Remove from operation queue */
    u->ops = u->ops->next;
    o->next = NULL;
    /* Resume evaluation */
    grind_booleans (u, o);
  }
  else
    conlog (t, "MATCHES received with no queries outstanding");
}

/* The answer is here. */
void
q_answer (struct conlist *t)
{
  struct resqueue *u;
  struct opqueue *o;
  u = find_resource (t);
  if (!u)
    panic ("Non-existent resource in q_answer", 1);
  o = u->ops;
  if (o)
  {
    if (o->stack)
      conlog (t, "Answer received, but there's stuff on the stack");
    else
    {
      /* Response datum is t->datum, query datum is on opqueue */
      /* Get rid of class line in query */
      while (o->datum)
      {
        if (strncmp (startoftext (o->datum), "Class:", 6))
        {
          t->enddatum->next = o->datum;
          t->enddatum = o->datum;
          o->datum = o->datum->next;
          t->enddatum->next = NULL;
        }
        else
        {
          struct dobj *u;
          u = o->datum->next;
          free (o->datum->text);
          free (o->datum);
          o->datum = u;
        }
      }
      /* If the resource didn't classify, then assign the class of the
       * resource.
       */
      if (u->restype)
      {
        char *class;
        class = classify (&(t->datum), NULL);
        if (class)
        {
          if (!strcmp (class, "U"))
            change_class (&(t->datum), u->restype);
        }
        else
          change_class (&(t->datum), u->restype);
      }
      /* Hooray!  It's done! */
      rebound (t, NULL);
    }
    /* Remove from operation queue */
    u->ops = u->ops->next;
    o->next = NULL;
    o->datum = o->query = NULL;
    free_opqueue (o);
  }
  else
    conlog (t, "Received answer, nothing in opqueue");
}

/* The code: (bits)
 *   1 = last item in message
 *   2 = start of data block
 *   4 = question
 */
void
res_datumline (struct conlist *t, char *buf)
{
  struct dobj *u;
  int flag, junk1;

  if (sscanf (buf, "%d", &flag) != 1)
    panic ("Non-datumline sent to res_datumline", 1);

  /* When we get the header for a block of binary data, we "sidetrack" it
   * until the block is received.  Then it comes back later along with the
   * data.
   */
  if ((flag&2)&&(!(t->bindata)))
  {
    long length;
    if (sscanf (buf, "%d Binary data follows ... %ld bytes", &junk1, &length)
    != 2)
    {
      conlog (t, "Unrecognized bindata header from");
      return;
    }
    t->binsize = length;
    t->sidetrack = strclone (buf);
    return;
  }

  /* Construct the datum from its pieces. */

  u = (struct dobj *) malloc (sizeof (struct dobj));
  if (!u)
    panic ("Couldn't malloc in datumline", 1);
  u->next = (struct dobj *)0;
  u->text = strclone (buf);
  u->binlen = t->binsize;
  u->bindata = t->bindata;
  t->binsize = 0;
  t->bindata = (char *)0;
  if (!(t->datum))
    t->datum = t->enddatum = u;
  else
  {
    /* Consing onto the end of a list =|-) */
    t->enddatum->next = u;
    t->enddatum = u;
  }

  if (!(flag&1))
    return;

  /* We have the complete message now */
  q_answer (t);

  free_dobj (t->datum); t->datum = t->enddatum = NULL;
  t->dmes = garbage;
}

/*   CONLOG something              Cause Unetd to write something into log
 *   READY                         Response when RESET is complete (no queries
 *                                 will be sent until READY)
 *   DECLASS something             Cause Unetd to declare some class
 *   MYTYPE something              Unetd will only deliver queries that match
 *                                 the specified class, which is declared
 *   BYE                           Resource going down.  Important.
 *   MATCHES [list of OID's]       These objects match the query.  MATCHES with
 *                                 no list of OID's means no matches.
 *   OOPS                          Failure response to a FETCH.
 *
 *   Successful responses to FETCH are sent in datum format.
 */
void
resourcehandler (char *buf, struct conlist *t, struct conlist **cons)
{
  if (strncmp (buf, "CONLOG ", 7) == 0)
    conlog (t, buf+7);
  else if (strncmp (buf, "READY", 5) == 0)
  {
    struct resqueue *u;
    u = find_resource (t);
    if (u)
    {
      u->ready = 1;
      conlog (t, "Resource ready");
    }
  }
  else if (strncmp (buf, "BYE", 3) == 0)
    delmark (t);
  else if (strncmp (buf, "DECLASS ", 8) == 0)
    declare_class (buf+8);
  else if (strncmp (buf, "MYTYPE ", 7) == 0)
  {
    struct resqueue *u;
    char temp[100];
    u = find_resource (t);
    declare_class (buf+7);
    if (!u)
      panic ("Failed to find resource", 1);
    if (u->restype)
      free (u->restype);
    u->restype = strclone (buf+7);
    sprintf (temp, "Declared type '%s'", u->restype);
    conlog (t, temp);
  }

  else if (strncmp (buf, "MATCHES", 7) == 0)
    grind_query (buf, t);

  else if (strncmp (buf, "BINDATA", 7) == 0)
  {
    /* BINDATA just flags the fact that we received some binary data.  It is
     * necessary so that the main program knows that something is going on.
     */
    ;
  }
  else if (strncmp (buf, "BIN_FINISH", 10) == 0)
  {
    /* BIN_FINISH means that the binary data object is ready to be snarfed. */
    res_datumline (t, t->sidetrack);
    free (t->sidetrack);
    t->sidetrack = NULL;
  }

  else if (strncmp (buf, "OOPS", 4) == 0)
  {
    struct resqueue *u;
    struct opqueue *o;
    conlog (t, "OOPS");
    u = find_resource (t);
    if (u)
    {
      o = u->ops;
      if (!o)
        conlog (t, "Nothing pending for OOPS");
      else
      {
        u->ops = u->ops->next;
        nextres (u, o);
      }
    }
    else
      conlog (t, "No such resource?!?");
  }

  else if (buf[0] >= '0' && buf[0] <= '9')
    res_datumline (t, buf);

  else
  {
    conlog (t, "Leaky pipe:");
    printf ("Data received:  '%s'\n", buf);
    fflush (stdout);
  }
}

/* This is where the queries come in. */
void
resource_query_inbox (struct conlist *t, struct dobj *u, int beenhere)
{
  add_cache ();
  if (res_ping (t, u))          /* resource.c */
    return;
  /* The regular resources only get one look at the query; anything else is a
   * waste of time, even though they *could* have changed since then.
   */
  if (!beenhere)
  {
    struct opqueue *o;
    if (!t)
    {
      /* Reset external resources -- not implemented */
      return;
    }
    else
    {
      o = (struct opqueue *) safmalloc (sizeof (struct opqueue));
      o->next = NULL;
      o->datum = t->datum;
      t->datum = t->enddatum = NULL;
      o->stack = o->exp = NULL;
      o->query = u;
      nextres (NULL, o);
      return;
    }
  }

  if (res_forward (t, u))         /* resource.c */
    return;

  /* res_forward should only fail if we are completely isolated. */
  if (t)
    fail_query (&(t->datum));     /* query.c */
}
