#line 1 "strfuns.c.ansi"
/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* strfuns.c  Commonly needed string handling functions.
 *
 *  This has been split off from strutil.c so that resource providers can
 *  #include it if they so choose.  It is not a standalone program.
 *
 *  A function "panic" must be provided by the #includer.
 */

/* Just to make sure */
#include <string.h>

#ifndef namlen
#define namlen 256
#endif

#ifndef min
#define min(a,b) (a > b ? b : a)
#endif

/* Malloc with error checking. */
void *
safmalloc (bytes)  size_t bytes;
{
  void *a;
  a = (void *) malloc (bytes);
  if (!a)
    panic ("Can't malloc", 1);
  return a;
}

/* The following functions replace strcasecmp, tolower, and things like that
 * which are broken, missing, or renamed on every system I use!
 */

void
strdowncase (s)  char *s;
{
  while (*s)
  {
    if (*s < 0)
      panic ("illegal characters passed to strdowncase", 1);
    if (*s >= 'A' && *s <= 'Z')
      *s = *s - 'A' + 'a';
    s++;
  }
}

int
cstrncmp (a, b, n)  char *a; char *b; int n;
{
  int c, cmp;
  if (strlen (a) +1 < n || strlen (b) +1 < n)
    n = min (strlen (a) +1, strlen (b) +1);
  for (c=0;c<n;c++)
  {
    cmp = ((a[c] >= 'A' && a[c] <= 'Z') ? a[c] - 'A' + 'a' : a[c])
            -
          ((b[c] >= 'A' && b[c] <= 'Z') ? b[c] - 'A' + 'a' : b[c]);
    if (cmp)
      return cmp;
  }
  return 0;
}

/* Case-sensitive */
int
sstrncmp (a, b, n)  char *a; char *b; int n;
{
  int c, cmp;
  if (strlen (a) +1 < n || strlen (b) +1 < n)
    n = min (strlen (a) +1, strlen (b) +1);
  for (c=0;c<n;c++)
  {
    cmp = a[c] - b[c];
    if (cmp)
      return cmp;
  }
  return 0;
}

int
cstrcmp (a, b)  char *a; char *b;
{
  return cstrncmp (a, b, min (strlen (a) +1, strlen (b) +1));
}

char *
cstrstr (s1, s2)  char *s1; char *s2;
{
  int a, s1len, s2len;
  if (!s1 || !s2)
    return NULL;
  s1len = strlen (s1);    /* Important to convert value to int from unsigned */
  s2len = strlen (s2);    /* or else s1len < s2len gets fubarred             */
  for (a=0;a<=s1len-s2len;a++)
    if (!(cstrncmp (s1+a, s2, s2len)))
      return s1+a;
  return NULL;
}

/* Case-sensitive */
char *
sstrstr (s1, s2)  char *s1; char *s2;
{
  int a, s1len, s2len;
  if (!s1 || !s2)
    return NULL;
  s1len = strlen (s1);    /* Important to convert value to int from unsigned */
  s2len = strlen (s2);    /* or else s1len < s2len gets fubarred             */
  for (a=0;a<=s1len-s2len;a++)
    if (!(sstrncmp (s1+a, s2, s2len)))
      return s1+a;
  return NULL;
}

/* Get rid of trailing CR's and LF's */
char *
noeol (s)  char *s;
{
  char *a;
  a = &(s[strlen(s)]);
  while (a > s)
    if (*(a-1) == '\n' || *(a-1) == '\r')
      *(--a) = '\0';
    else
      break;
  return s;
}

/* Get rid of punctuation */
char *
nopunct (s)  char *s;
{
  char *a;
  a = s;
  while (*a)
  {
    if (ispunct (*a))
      *a = ' ';
    a++;
  }
  return s;
}

/* For some weird reason, strdup is not in ANSI C. */
char *
strclone (buf)  char *buf;
{
  char *t;
  if (buf)
  {
    t = (char *) safmalloc (strlen (buf) +1);
    strcpy (t, buf);
    return t;
  }
  else
    return NULL;
}

/* Find words in a text string. */
char *
strword (searchme, word)  char *searchme; char *word;
{
  char *t, *a, *q=NULL;
  t = strclone (searchme);
  a = strtok (t, " \n\r\t\v\f&|~()[]");
  while (a)
  {
    if (!(cstrcmp (a, word)))
    {
      q = searchme + (a - t);
      break;
    }
    a = strtok (NULL, " \n\r\t\v\f&|~()[]");
  }
  free (t);
  return q;
}

/* Return number of words in string2 that occur in string1.  Length is set
 * to the number of words in string2 if it's not NULL.
 */
int
wrdmatches (string1, string2, length)  char *string1; char *string2; int *length;
{
  char a[namlen];
  int t_degree=0;
  if (length)
    (*length) = 0;
  if ((!string1) || (!string2))
    return 0;

  /* Can't use strtok here because strword uses it. */
  /* This substitute requires string2 to have a fairly rigid structure. */

  while (*string2 == ' ')
    string2++;
  while (sscanf (string2, "%s", a) == 1)
  {
    if (length)
      (*length)++;
    if (strword (string1, a))
      t_degree++;
    string2 += strlen(a);
    while (*string2 == ' ')
      string2++;
  }
  return t_degree;
}

/* A variation on strncmp that is more sensitive to word breaks.
 *    Let s1 = "rec woodchucks" and s2 = "rec wood"
 *    strncmp (s1, s2, strlen(s2)) returns 0.
 *    struncump (s1, s2) returns 1 because wood is not the same as woodchucks.
 *
 *  WARNING  This function is now specialized for comparing classifications.
 *     The special case "U" universal class is now recognized.
 */
int
struncump (s1, s2)  char *s1; char *s2;
{
  if (!s2 || !s1)
    return 0;
  if ((!(strcmp (s1, "U"))) || (!(strcmp (s2, "U"))))
    return 0;
  if (cstrncmp (s1, s2, strlen (s2)))
    return 1;
  if (strlen (s2) >= strlen (s1))
    return 0;
  if (s1[strlen(s2)] != ' ')
    return 1;
  return 0;
}
