/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "alibi.h"

#include "strfuns.c"

/* Add data to the back of a queue */
void
append_to_queue (char **buff, long *length, char *newdata, long newlength, long *memlength)
{
  if (*buff == NULL)
  {
    if (*length)
      panic ("Null buffer with nonzero length in append_to_queue", 1);
    *buff = (char *) safmalloc (newlength);
    *length = newlength;
    memcpy (*buff, newdata, newlength);
  }
  else
  {
    *length += newlength;
    if (*memlength < *length)
    {
      /* Doing lots of reallocs seems to be a bad thing.  This should reduce
       * the number of reallocs.
       */
      *memlength = *length << 1;
      if ((*buff = (char *) realloc (*buff, *memlength)) == NULL)
        panic ("Can't realloc bigger in append_to_queue", 1);
    }
    memcpy (*buff + *length - newlength, newdata, newlength);
  }
}

/* Remove data from the front of a queue */
void
shorten_queue (char **buff, long *length, long remove_amount, long *memlength)
{
  if (remove_amount > *length)
    panic ("Underflow in shorten_queue", 1);
  *length -= remove_amount;
  if (*length)
  {
    memcpy (*buff, (*buff)+remove_amount, *length);
    /* Avoid excess reallocs */
#if 0
    if ((*buff = (char *) realloc (*buff, *length * sizeof (char))) == NULL)
        panic ("Can't realloc smaller in shorten_queue", 1);
#endif
  }
  else
  {
    free (*buff);
    *buff = NULL;
    *memlength = 0;
  }
}

void
free_dobj (struct dobj *t)
{
  while (t)
  {
    struct dobj *u;
    u = t->next;
    if (t->text)
      free (t->text);
    if (t->bindata)
      free (t->bindata);
    free (t);
    t = u;
  }
}

char *
tstartoftext (char *u)
{
  char *a;
  if (!u)
    return NULL;
  a = strchr (u, ' ');
  if (!a)
    return (u + strlen (u));
  while (*a == ' ')
    a++;
  return a;
}

/* Given a line of a message, this returns a pointer to the start of the text,
 * after the flag or keyword or whatever.
 */
char *
startoftext (struct dobj *u)
{
  if (!u)
    panic ("No u in startoftext", 1);
  if (!(u->text))
    panic ("No text in startoftext", 1);
  return tstartoftext (u->text);
}

/* Remove irrelevant terms from a text string.  Returns number removed. */
int
sremove (char *s, char *delme)
{
  char *t;
  int a, ld, count=0;
  ld = strlen (delme);
  if (!ld)
    return 0;
  t = strword (s, delme);
  while (t)
  {
    count++;
    for (a=0;a<strlen(t+ld)+1;a++)
      t[a] = t[a+ld];
    t = strword (s, delme);
  }
  return count;
}

/* Multiple sremoves. */
void
sremall (char *s, char *delme)
{
  char smallbuf[80];
  while (sscanf (delme, "%s", smallbuf) == 1)
  {
    delme = cstrstr(delme, smallbuf) + strlen(smallbuf);
    sremove (s, smallbuf);
  }
}

struct dobj *
clone_dobj (struct dobj *t)
{
  struct dobj *u, *head=NULL, *tail=NULL;
  while (t)
  {
    u = (struct dobj *) malloc (sizeof (struct dobj));
    if (!u)
      panic ("Can't malloc in clone_dobj", 1);
    u->text = strclone (t->text);
    u->bindata = (char *) malloc (t->binlen);
    if (!(u->bindata))
      panic ("Can't malloc in clone_dobj", 1);
    memcpy (u->bindata, t->bindata, t->binlen);
    u->binlen = t->binlen;
    u->next = NULL;
    if (head)
    {
      tail->next = u;
      tail = u;
    }
    else
      head = tail = u;
    t = t->next;
  }
  return head;
}

/* This is NOT just a cloning function.  It filters out all fields not of
 * interest to the cache and adds an informational line.
 */
struct dobj *
filter_clone_dobj (struct dobj *t, time_t now)
{
  struct dobj *u, *head=NULL, *tail=NULL;
  char smalbuf[80];
  head = tail = (struct dobj *) malloc (sizeof (struct dobj));
  if (!head)
    panic ("Can't malloc in filter_clone_dobj", 1);
  sprintf (smalbuf, "0 This datum has been in cache since %.24s",
    ctime(&now));
  head->text = strclone (smalbuf);
  head->bindata = NULL;
  head->binlen = 0l;
  head->next = NULL;
  while (t)
  {
    int flag, junk;
    unsigned long junk2;
    char cjunk[namlen];
    if (sscanf (t->text, "%d", &flag) == 1)
    {
      if ((flag && strcmp (t->text, "1 Have a nice day =|-)"))
        || ((!flag) &&
        sscanf (t->text, "%d Query path: %s", &junk, cjunk) != 2 &&
        sscanf (t->text, "%d Answer for %lu", &junk, &junk2) != 2 &&
        sscanf (t->text, "%d Start time: %lu", &junk, &junk2) != 2 &&
        sscanf (t->text, "%d Response path:%c", &junk, cjunk) != 2))
      {
        u = (struct dobj *) malloc (sizeof (struct dobj));
        if (!u)
          panic ("Can't malloc in filter_clone_dobj", 1);
        u->text = strclone (t->text);
        if (t->binlen)
        {
          u->bindata = (char *) malloc (t->binlen);
          if (!(u->bindata))
            panic ("Can't malloc in filter_clone_dobj", 1);
          memcpy (u->bindata, t->bindata, t->binlen);
        }
        else
          u->bindata = NULL;
        u->binlen = t->binlen;
        u->next = NULL;
        tail->next = u;
        tail = u;
      }
    }
    t = t->next;
  }
  return head;
}
