/* -*-text-*-
 *
 *   cron_rc.c    Cron program to restart daemons after a reboot.
 *   DWF  3/8/94
 *
 *   This is a substitute for editing rc.local if your system administrator
 *   won't allow it.  The crontab entry should look like this:
 *
 *        0,15,30,45 * * * * /whatever/pathname/cron_rc
 *
 *   After a reboot, this program will attempt to execute a shell script
 *   in your home directory with the same name as the host.  Run this
 *   program with a command-line argument, and it will print out the full
 *   pathname of the script it expects to find.  This script is doubly
 *   important since its last modification time is "touched" by cron_rc
 *   to remember the previous reboot time.
 *
 *   Sun:    -lrpcsvc [-lnsl]        ; -lnsl needed on SYSV Suns
 *   SGI:    -lrpcsvc -lsun
 *
 *   By default, this program uses RPC to determine the uptime.  This is the
 *   most portable way to do it because RPC responses have to be standardized
 *   in order to work at all.  If your machine does not support this RPC call
 *   (try 'rup localhost'), do ls -al /dev/console and see whether the last
 *   modification time is close to when the system was last rebooted.  If it
 *   is, you can #define use_dev_console_hack (see below) to get this program
 *   to work and leave out all the RPC libraries.
 *
 *   The HP-UX compatibility hacks work only with use_dev_console_hack.
 */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#undef use_dev_console_hack

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef use_dev_console_hack
#include <rpc/rpc.h>
#include <rpcsvc/rstat.h>
#endif

#ifdef __hpux
#include <utime.h>
#endif

#ifdef __hpux
struct utimbuf tvp;
time_t boottime;
#else
struct timeval boottime, tvp[2];
#endif
char fname[256];

void
get_fname ()
{
  char hostname[256], *home;
  if (gethostname (hostname, 256))
  {
    perror ("gethostname");
    exit (-1);
  }
  if (!(home = getenv ("HOME")))
  {
    puts ("No HOME environment variable");
    exit (-1);
  }
  sprintf (fname, "%s/%s", home, hostname);
}

void
get_boottime ()
{
#ifdef use_dev_console_hack
  struct stat s;
  if (stat ("/dev/console", &s) != -1)
  {
#ifdef __hpux
    tvp.actime = tvp.modtime = boottime = s.st_mtime;
#else
    tvp[0].tv_sec  = tvp[1].tv_sec  = boottime.tv_sec
      = s.st_mtime;
    tvp[0].tv_usec = tvp[1].tv_usec = boottime.tv_usec
      = 0;
#endif
  }
  else
  {
    perror ("stat");
    exit (-1);
  }
#else
  CLIENT *rstat_clnt;
  struct statstime host_stat;
  struct timeval timeout;
  timeout.tv_sec = 5;
  timeout.tv_usec = 0;
  if (!(rstat_clnt=clnt_create("localhost", RSTATPROG, RSTATVERS_TIME, "udp")))
  {
    puts ("RPC failure (clnt_create)");
    exit (-1);
  }
  if (clnt_call(rstat_clnt, RSTATPROC_STATS, xdr_void, NULL, xdr_statstime,
  &host_stat, timeout) != RPC_SUCCESS)
  {
    puts ("RPC failure (clnt_call)");
    exit (-1);
  }
  clnt_destroy(rstat_clnt);
  tvp[0].tv_sec  = tvp[1].tv_sec  = boottime.tv_sec
    = host_stat.boottime.tv_sec;
  tvp[0].tv_usec = tvp[1].tv_usec = boottime.tv_usec
    = host_stat.boottime.tv_usec;
#endif
}

void
touch_script ()
{
#ifdef __hpux
  if (utime  (fname, &tvp))
#else
  if (utimes (fname, tvp))
#endif
    perror ("utimes");
}

main (argc, argv)
int argc;
char **argv;
{
  struct stat s;
  get_fname ();
  if (argc != 1)
  {
    printf ("Will use the script called %s\n", fname);
    exit (0);
  }
  get_boottime ();
  if (stat (fname, &s) != -1)
  {
    /* Allow for drifting of clocks */
#ifdef __hpux
    if (s.st_mtime < boottime - 60)
#else
    if (s.st_mtime < boottime.tv_sec - 60)
#endif
      system (fname);
#ifdef __hpux
    if (s.st_mtime != boottime)
#else
    if (s.st_mtime != boottime.tv_sec)
#endif
      touch_script ();
  }
  else
    perror ("stat");
  return 0;
}
